from typing import Optional

from pymobiledevice3.exceptions import DeprecationError, PyMobileDevice3Exception
from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.lockdown_service_provider import LockdownServiceProvider
from pymobiledevice3.services.lockdown_service import LockdownService

MobileGestaltKeys = [
    "3GProximityCapability",
    "3GVeniceCapability",
    "3Gvenice",
    "3d-imagery",
    "3d-maps",
    "64-bit",
    "720p",
    "720pPlaybackCapability",
    "APNCapability",
    "ARM64ExecutionCapability",
    "ARMV6ExecutionCapability",
    "ARMV7ExecutionCapability",
    "ARMV7SExecutionCapability",
    "ASTC",
    "AWDID",
    "AWDLCapability",
    "AccelerometerCapability",
    "AccessibilityCapability",
    "AcousticID",
    "ActivationProtocol",
    "ActiveWirelessTechnology",
    "ActuatorResonantFrequency",
    "AdditionalTextTonesCapability",
    "AggregateDevicePhotoZoomFactor",
    "AggregateDeviceVideoZoomFactor",
    "AirDropCapability",
    "AirDropRestriction",
    "AirplaneMode",
    "AirplayMirroringCapability",
    "AllDeviceCapabilities",
    "Allow32BitApps",
    "AllowOnlyATVCPSDKApps",
    "AllowYouTube",
    "AllowYouTubePlugin",
    "AmbientLightSensorCapability",
    "AmbientLightSensorSerialNumber",
    "ApNonce",
    "ApNonceRetrieve",
    "AppCapacityTVOS",
    "AppStore",
    "AppStoreCapability",
    "AppleInternalInstallCapability",
    "AppleNeuralEngineSubtype",
    "ApplicationInstallationCapability",
    "ArcModuleSerialNumber",
    "ArrowChipID",
    "ArrowUniqueChipID",
    "ArtworkTraits",
    "AssistantCapability",
    "AudioPlaybackCapability",
    "AutoFocusCameraCapability",
    "AvailableDisplayZoomSizes",
    "BacklightCapability",
    "BasebandAPTimeSync",
    "BasebandBoardSnum",
    "BasebandCertId",
    "BasebandChipId",
    "BasebandChipset",
    "BasebandClass",
    "BasebandFirmwareManifestData",
    "BasebandFirmwareUpdateInfo",
    "BasebandFirmwareVersion",
    "BasebandKeyHashInformation",
    "BasebandPostponementStatus",
    "BasebandPostponementStatusBlob",
    "BasebandRegionSKU",
    "BasebandRegionSKURadioTechnology",
    "BasebandSecurityInfoBlob",
    "BasebandSerialNumber",
    "BasebandSkeyId",
    "BasebandStatus",
    "BasebandUniqueId",
    "BatteryCurrentCapacity",
    "BatteryIsCharging",
    "BatteryIsFullyCharged",
    "BatterySerialNumber",
    "BlueLightReductionSupported",
    "BluetoothAddress",
    "BluetoothAddressData",
    "BluetoothCapability",
    "BluetoothLE2Capability",
    "BluetoothLECapability",
    "BoardId",
    "BoardRevision",
    "BootManifestHash",
    "BootNonce",
    "BridgeBuild",
    "BridgeRestoreVersion",
    "BuddyLanguagesAnimationRequiresOptimization",
    "BuildID",
    "BuildVersion",
    "C2KDeviceCapability",
    "CPUArchitecture",
    "CPUSubType",
    "CPUType",
    "CallForwardingCapability",
    "CallWaitingCapability",
    "CallerIDCapability",
    "CameraAppUIVersion",
    "CameraCapability",
    "CameraFlashCapability",
    "CameraFrontFlashCapability",
    "CameraHDR2Capability",
    "CameraHDRVersion",
    "CameraLiveEffectsCapability",
    "CameraMaxBurstLength",
    "CameraRestriction",
    "CarrierBundleInfoArray",
    "CarrierInstallCapability",
    "CellBroadcastCapability",
    "CellularDataCapability",
    "CellularTelephonyCapability",
    "CertificateProductionStatus",
    "CertificateSecurityMode",
    "ChipID",
    "CloudPhotoLibraryCapability",
    "CoastlineGlowRenderingCapability",
    "CompassCalibration",
    "CompassCalibrationDictionary",
    "CompassType",
    "ComputerName",
    "ConferenceCallType",
    "ConfigNumber",
    "ContainsCellularRadioCapability",
    "ContinuityCapability",
    "CoreRoutineCapability",
    "CoverglassSerialNumber",
    "DMin",
    "DataPlanCapability",
    "DebugBoardRevision",
    "DelaySleepForHeadsetClickCapability",
    "DesenseBuild",
    "DeviceAlwaysPrewarmActuator",
    "DeviceBackGlassMaterial",
    "DeviceBackingColor",
    "DeviceBrand",
    "DeviceClass",
    "DeviceClassNumber",
    "DeviceColor",
    "DeviceColorMapPolicy",
    "DeviceCornerRadius",
    "DeviceCoverGlassColor",
    "DeviceCoverGlassMaterial",
    "DeviceCoverMaterial",
    "DeviceEnclosureColor",
    "DeviceEnclosureMaterial",
    "DeviceEnclosureRGBColor",
    "DeviceHasAggregateCamera",
    "DeviceHousingColor",
    "DeviceIsMuseCapable",
    "DeviceKeyboardCalibration",
    "DeviceLaunchTimeLimitScale",
    "DeviceName",
    "DeviceNameString",
    "DevicePrefers3DBuildingStrokes",
    "DevicePrefersBuildingStrokes",
    "DevicePrefersCheapTrafficShaders",
    "DevicePrefersProceduralAntiAliasing",
    "DevicePrefersTrafficAlpha",
    "DeviceProximityCapability",
    "DeviceRGBColor",
    "DeviceRequiresPetalOptimization",
    "DeviceRequiresProximityAmeliorations",
    "DeviceRequiresSoftwareBrightnessCalculations",
    "DeviceSceneUpdateTimeLimitScale",
    "DeviceSubBrand",
    "DeviceSupports1080p",
    "DeviceSupports3DImagery",
    "DeviceSupports3DMaps",
    "DeviceSupports3rdPartyHaptics",
    "DeviceSupports4G",
    "DeviceSupports4k",
    "DeviceSupports64Bit",
    "DeviceSupports720p",
    "DeviceSupports9Pin",
    "DeviceSupportsAOP",
    "DeviceSupportsARKit",
    "DeviceSupportsASTC",
    "DeviceSupportsAdaptiveMapsUI",
    "DeviceSupportsAlwaysListening",
    "DeviceSupportsAlwaysOnCompass",
    "DeviceSupportsAlwaysOnTime",
    "DeviceSupportsApplePencil",
    "DeviceSupportsAutoLowLightVideo",
    "DeviceSupportsAvatars",
    "DeviceSupportsBatteryModuleAuthentication",
    "DeviceSupportsBerkelium2",
    "DeviceSupportsCCK",
    "DeviceSupportsCameraCaptureOnTouchDown",
    "DeviceSupportsCameraDeferredProcessing",
    "DeviceSupportsCameraHaptics",
    "DeviceSupportsCarIntegration",
    "DeviceSupportsCinnamon",
    "DeviceSupportsClosedLoopHaptics",
    "DeviceSupportsCrudeProx",
    "DeviceSupportsDClr",
    "DeviceSupportsDoNotDisturbWhileDriving",
    "DeviceSupportsELabel",
    "DeviceSupportsEnhancedAC3",
    "DeviceSupportsEnvironmentalDosimetry",
    "DeviceSupportsExternalHDR",
    "DeviceSupportsFloorCounting",
    "DeviceSupportsHDRDeferredProcessing",
    "DeviceSupportsHMEInARKit",
    "DeviceSupportsHaptics",
    "DeviceSupportsHardwareDetents",
    "DeviceSupportsHeartHealthAlerts",
    "DeviceSupportsHeartRateVariability",
    "DeviceSupportsHiResBuildings",
    "DeviceSupportsLineIn",
    "DeviceSupportsLiquidDetection_CorrosionMitigation",
    "DeviceSupportsLivePhotoAuto",
    "DeviceSupportsLongFormAudio",
    "DeviceSupportsMapsBlurredUI",
    "DeviceSupportsMapsOpticalHeading",
    "DeviceSupportsMomentCapture",
    "DeviceSupportsNFC",
    "DeviceSupportsNavigation",
    "DeviceSupportsNewton",
    "DeviceSupportsOnDemandPhotoAnalysis",
    "DeviceSupportsP3ColorspaceVideoRecording",
    "DeviceSupportsPeriodicALSUpdates",
    "DeviceSupportsPhotosLocalLight",
    "DeviceSupportsPortraitIntensityAdjustments",
    "DeviceSupportsPortraitLightEffectFilters",
    "DeviceSupportsRGB10",
    "DeviceSupportsRaiseToSpeak",
    "DeviceSupportsSiDP",
    "DeviceSupportsSideButtonClickSpeed",
    "DeviceSupportsSimplisticRoadMesh",
    "DeviceSupportsSingleCameraPortrait",
    "DeviceSupportsSiriBargeIn",
    "DeviceSupportsSiriSpeaks",
    "DeviceSupportsSiriSpokenMessages",
    "DeviceSupportsSpatialOverCapture",
    "DeviceSupportsStereoAudioRecording",
    "DeviceSupportsStudioLightPortraitPreview",
    "DeviceSupportsSwimmingWorkouts",
    "DeviceSupportsTapToWake",
    "DeviceSupportsTelephonyOverUSB",
    "DeviceSupportsTethering",
    "DeviceSupportsToneMapping",
    "DeviceSupportsUSBTypeC",
    "DeviceSupportsVSHCompensation",
    "DeviceSupportsVoiceOverCanUseSiriVoice",
    "DeviceSupportsWebkit",
    "DeviceSupportsWirelessSplitting",
    "DeviceSupportsYCbCr10",
    "DeviceVariant",
    "DeviceVariantGuess",
    "DiagData",
    "DictationCapability",
    "DieId",
    "DiskUsage",
    "DisplayDriverICChipID",
    "DisplayFCCLogosViaSoftwareCapability",
    "DisplayMirroringCapability",
    "DisplayPortCapability",
    "DualSIMActivationPolicyCapable",
    "EUICCChipID",
    "EffectiveProductionStatus",
    "EffectiveProductionStatusAp",
    "EffectiveProductionStatusSEP",
    "EffectiveSecurityMode",
    "EffectiveSecurityModeAp",
    "EffectiveSecurityModeSEP",
    "EncodeAACCapability",
    "EncryptedDataPartitionCapability",
    "EnforceCameraShutterClick",
    "EnforceGoogleMail",
    "EthernetMacAddress",
    "EthernetMacAddressData",
    "ExplicitContentRestriction",
    "ExternalChargeCapability",
    "ExternalPowerSourceConnected",
    "FDRSealingStatus",
    "FMFAllowed",
    "FaceTimeBackCameraTemporalNoiseReductionMode",
    "FaceTimeBitRate2G",
    "FaceTimeBitRate3G",
    "FaceTimeBitRateLTE",
    "FaceTimeBitRateWiFi",
    "FaceTimeCameraRequiresFastSwitchOptions",
    "FaceTimeCameraSupportsHardwareFaceDetection",
    "FaceTimeDecodings",
    "FaceTimeEncodings",
    "FaceTimeFrontCameraTemporalNoiseReductionMode",
    "FaceTimePhotosOptIn",
    "FaceTimePreferredDecoding",
    "FaceTimePreferredEncoding",
    "FirmwareNonce",
    "FirmwarePreflightInfo",
    "FirmwareVersion",
    "FirstPartyLaunchTimeLimitScale",
    "ForwardCameraCapability",
    "FrontCameraOffsetFromDisplayCenter",
    "FrontCameraRotationFromDisplayNormal",
    "FrontFacingCameraAutoHDRCapability",
    "FrontFacingCameraBurstCapability",
    "FrontFacingCameraCapability",
    "FrontFacingCameraHDRCapability",
    "FrontFacingCameraHDROnCapability",
    "FrontFacingCameraHFRCapability",
    "FrontFacingCameraHFRVideoCapture1080pMaxFPS",
    "FrontFacingCameraHFRVideoCapture720pMaxFPS",
    "FrontFacingCameraMaxVideoZoomFactor",
    "FrontFacingCameraModuleSerialNumber",
    "FrontFacingCameraStillDurationForBurst",
    "FrontFacingCameraVideoCapture1080pMaxFPS",
    "FrontFacingCameraVideoCapture4kMaxFPS",
    "FrontFacingCameraVideoCapture720pMaxFPS",
    "FrontFacingIRCameraModuleSerialNumber",
    "FrontFacingIRStructuredLightProjectorModuleSerialNumber",
    "Full6FeaturesCapability",
    "GPSCapability",
    "GSDeviceName",
    "GameKitCapability",
    "GasGaugeBatteryCapability",
    "GreenTeaDeviceCapability",
    "GyroscopeCapability",
    "H264EncoderCapability",
    "HDRImageCaptureCapability",
    "HDVideoCaptureCapability",
    "HEVCDecoder10bitSupported",
    "HEVCDecoder12bitSupported",
    "HEVCDecoder8bitSupported",
    "HEVCEncodingCapability",
    "HMERefreshRateInARKit",
    "HWModelStr",
    "HallEffectSensorCapability",
    "HardwareEncodeSnapshotsCapability",
    "HardwareKeyboardCapability",
    "HardwarePlatform",
    "HardwareSnapshotsRequirePurpleGfxCapability",
    "HasAllFeaturesCapability",
    "HasAppleNeuralEngine",
    "HasBaseband",
    "HasBattery",
    "HasDaliMode",
    "HasExtendedColorDisplay",
    "HasIcefall",
    "HasInternalSettingsBundle",
    "HasMesa",
    "HasPKA",
    "HasSEP",
    "HasSpringBoard",
    "HasThinBezel",
    "HealthKitCapability",
    "HearingAidAudioEqualizationCapability",
    "HearingAidLowEnergyAudioCapability",
    "HearingAidPowerReductionCapability",
    "HiDPICapability",
    "HiccoughInterval",
    "HideNonDefaultApplicationsCapability",
    "HighestSupportedVideoMode",
    "HomeButtonType",
    "HomeScreenWallpaperCapability",
    "IDAMCapability",
    "IOSurfaceBackedImagesCapability",
    "IOSurfaceFormatDictionary",
    "IceFallID",
    "IcefallInRestrictedMode",
    "IcefallInfo",
    "Image4CryptoHashMethod",
    "Image4Supported",
    "InDiagnosticsMode",
    "IntegratedCircuitCardIdentifier",
    "IntegratedCircuitCardIdentifier2",
    "InternalBuild",
    "InternationalMobileEquipmentIdentity",
    "InternationalMobileEquipmentIdentity2",
    "InternationalSettingsCapability",
    "InverseDeviceID",
    "IsEmulatedDevice",
    "IsLargeFormatPhone",
    "IsPwrOpposedVol",
    "IsServicePart",
    "IsSimulator",
    "IsThereEnoughBatteryLevelForSoftwareUpdate",
    "IsUIBuild",
    "JasperSerialNumber",
    "LTEDeviceCapability",
    "LaunchTimeLimitScaleSupported",
    "LisaCapability",
    "LoadThumbnailsWhileScrollingCapability",
    "LocalizedDeviceNameString",
    "LocationRemindersCapability",
    "LocationServicesCapability",
    "LowPowerWalletMode",
    "LunaFlexSerialNumber",
    "LynxPublicKey",
    "LynxSerialNumber",
    "MLBSerialNumber",
    "MLEHW",
    "MMSCapability",
    "MacBridgingKeys",
    "MagnetometerCapability",
    "MainDisplayRotation",
    "MainScreenCanvasSizes",
    "MainScreenClass",
    "MainScreenHeight",
    "MainScreenOrientation",
    "MainScreenPitch",
    "MainScreenScale",
    "MainScreenStaticInfo",
    "MainScreenWidth",
    "MarketingNameString",
    "MarketingProductName",
    "MarketingVersion",
    "MaxH264PlaybackLevel",
    "MaximumScreenScale",
    "MedusaFloatingLiveAppCapability",
    "MedusaOverlayAppCapability",
    "MedusaPIPCapability",
    "MedusaPinnedAppCapability",
    "MesaSerialNumber",
    "MetalCapability",
    "MicrophoneCapability",
    "MicrophoneCount",
    "MinimumSupportediTunesVersion",
    "MixAndMatchPrevention",
    "MobileDeviceMinimumVersion",
    "MobileEquipmentIdentifier",
    "MobileEquipmentInfoBaseId",
    "MobileEquipmentInfoBaseProfile",
    "MobileEquipmentInfoBaseVersion",
    "MobileEquipmentInfoCSN",
    "MobileEquipmentInfoDisplayCSN",
    "MobileSubscriberCountryCode",
    "MobileSubscriberNetworkCode",
    "MobileWifi",
    "ModelNumber",
    "MonarchLowEndHardware",
    "MultiLynxPublicKeyArray",
    "MultiLynxSerialNumberArray",
    "MultitaskingCapability",
    "MultitaskingGesturesCapability",
    "MusicStore",
    "MusicStoreCapability",
    "N78aHack",
    "NFCRadio",
    "NFCRadioCalibrationDataPresent",
    "NFCUniqueChipID",
    "NVRAMDictionary",
    "NandControllerUID",
    "NavajoFusingState",
    "NikeIpodCapability",
    "NotGreenTeaDeviceCapability",
    "OLEDDisplay",
    "OTAActivationCapability",
    "OfflineDictationCapability",
    "OpenGLES1Capability",
    "OpenGLES2Capability",
    "OpenGLES3Capability",
    "OpenGLESVersion",
    "PTPLargeFilesCapability",
    "PanelSerialNumber",
    "PanoramaCameraCapability",
    "PartitionType",
    "PasswordConfigured",
    "PasswordProtected",
    "PearlCameraCapability",
    "PearlIDCapability",
    "PeekUICapability",
    "PeekUIWidth",
    "Peer2PeerCapability",
    "PersonalHotspotCapability",
    "PhoneNumber",
    "PhoneNumber2",
    "PhosphorusCapability",
    "PhotoAdjustmentsCapability",
    "PhotoCapability",
    "PhotoSharingCapability",
    "PhotoStreamCapability",
    "PhotosPostEffectsCapability",
    "PiezoClickerCapability",
    "PintoMacAddress",
    "PintoMacAddressData",
    "PipelinedStillImageProcessingCapability",
    "PlatformStandAloneContactsCapability",
    "PlatinumCapability",
    "ProductHash",
    "ProductName",
    "ProductType",
    "ProductVersion",
    "ProximitySensorCalibration",
    "ProximitySensorCalibrationDictionary",
    "ProximitySensorCapability",
    "RF-exposure-separation-distance",
    "RFExposureSeparationDistance",
    "RawPanelSerialNumber",
    "RearCameraCapability",
    "RearCameraOffsetFromDisplayCenter",
    "RearFacingCamera60fpsVideoCaptureCapability",
    "RearFacingCameraAutoHDRCapability",
    "RearFacingCameraBurstCapability",
    "RearFacingCameraCapability",
    "RearFacingCameraHDRCapability",
    "RearFacingCameraHDROnCapability",
    "RearFacingCameraHFRCapability",
    "RearFacingCameraHFRVideoCapture1080pMaxFPS",
    "RearFacingCameraHFRVideoCapture720pMaxFPS",
    "RearFacingCameraMaxVideoZoomFactor",
    "RearFacingCameraModuleSerialNumber",
    "RearFacingCameraStillDurationForBurst",
    "RearFacingCameraSuperWideCameraCapability",
    "RearFacingCameraTimeOfFlightCameraCapability",
    "RearFacingCameraVideoCapture1080pMaxFPS",
    "RearFacingCameraVideoCapture4kMaxFPS",
    "RearFacingCameraVideoCapture720pMaxFPS",
    "RearFacingCameraVideoCaptureFPS",
    "RearFacingLowLightCameraCapability",
    "RearFacingSuperWideCameraModuleSerialNumber",
    "RearFacingTelephotoCameraCapability",
    "RearFacingTelephotoCameraModuleSerialNumber",
    "RecoveryOSVersion",
    "RegionCode",
    "RegionInfo",
    "RegionSupportsCinnamon",
    "RegionalBehaviorAll",
    "RegionalBehaviorChinaBrick",
    "RegionalBehaviorEUVolumeLimit",
    "RegionalBehaviorGB18030",
    "RegionalBehaviorGoogleMail",
    "RegionalBehaviorNTSC",
    "RegionalBehaviorNoPasscodeLocationTiles",
    "RegionalBehaviorNoVOIP",
    "RegionalBehaviorNoWiFi",
    "RegionalBehaviorShutterClick",
    "RegionalBehaviorValid",
    "RegionalBehaviorVolumeLimit",
    "RegulatoryModelNumber",
    "ReleaseType",
    "RemoteBluetoothAddress",
    "RemoteBluetoothAddressData",
    "RenderWideGamutImagesAtDisplayTime",
    "RendersLetterPressSlowly",
    "RequiredBatteryLevelForSoftwareUpdate",
    "RestoreOSBuild",
    "RestrictedCountryCodes",
    "RingerSwitchCapability",
    "RosalineSerialNumber",
    "RoswellChipID",
    "RotateToWakeStatus",
    "SBAllowSensitiveUI",
    "SBCanForceDebuggingInfo",
    "SDIOManufacturerTuple",
    "SDIOProductInfo",
    "SEInfo",
    "SEPNonce",
    "SIMCapability",
    "SIMPhonebookCapability",
    "SIMStatus",
    "SIMStatus2",
    "SIMTrayStatus",
    "SIMTrayStatus2",
    "SMSCapability",
    "SavageChipID",
    "SavageInfo",
    "SavageSerialNumber",
    "SavageUID",
    "ScreenDimensions",
    "ScreenDimensionsCapability",
    "ScreenRecorderCapability",
    "ScreenSerialNumber",
    "SecondaryBluetoothMacAddress",
    "SecondaryBluetoothMacAddressData",
    "SecondaryEthernetMacAddress",
    "SecondaryEthernetMacAddressData",
    "SecondaryWifiMacAddress",
    "SecondaryWifiMacAddressData",
    "SecureElement",
    "SecureElementID",
    "SecurityDomain",
    "SensitiveUICapability",
    "SerialNumber",
    "ShoeboxCapability",
    "ShouldHactivate",
    "SiKACapability",
    "SigningFuse",
    "SiliconBringupBoard",
    "SimultaneousCallAndDataCurrentlySupported",
    "SimultaneousCallAndDataSupported",
    "SiriGestureCapability",
    "SiriOfflineCapability",
    "Skey",
    "SoftwareBehavior",
    "SoftwareBundleVersion",
    "SoftwareDimmingAlpha",
    "SpeakerCalibrationMiGa",
    "SpeakerCalibrationSpGa",
    "SpeakerCalibrationSpTS",
    "SphereCapability",
    "StarkCapability",
    "StockholmJcopInfo",
    "StrictWakeKeyboardCases",
    "SupportedDeviceFamilies",
    "SupportedKeyboards",
    "SupportsBurninMitigation",
    "SupportsEDUMU",
    "SupportsForceTouch",
    "SupportsIrisCapture",
    "SupportsLowPowerMode",
    "SupportsPerseus",
    "SupportsRotateToWake",
    "SupportsSOS",
    "SupportsSSHBButtonType",
    "SupportsTouchRemote",
    "SysCfg",
    "SysCfgDict",
    "SystemImageID",
    "SystemTelephonyOfAnyKindCapability",
    "TVOutCrossfadeCapability",
    "TVOutSettingsCapability",
    "TelephonyCapability",
    "TelephonyMaximumGeneration",
    "TimeSyncCapability",
    "TopModuleAuthChipID",
    "TouchDelivery120Hz",
    "TouchIDCapability",
    "TristarID",
    "UIBackgroundQuality",
    "UIParallaxCapability",
    "UIProceduralWallpaperCapability",
    "UIReachability",
    "UMTSDeviceCapability",
    "UnifiedIPodCapability",
    "UniqueChipID",
    "UniqueDeviceID",
    "UniqueDeviceIDData",
    "UserAssignedDeviceName",
    "UserIntentPhysicalButtonCGRect",
    "UserIntentPhysicalButtonCGRectString",
    "UserIntentPhysicalButtonNormalizedCGRect",
    "VOIPCapability",
    "VeniceCapability",
    "VibratorCapability",
    "VideoCameraCapability",
    "VideoStillsCapability",
    "VoiceControlCapability",
    "VolumeButtonCapability",
    "WAGraphicQuality",
    "WAPICapability",
    "WLANBkgScanCache",
    "WSKU",
    "WatchCompanionCapability",
    "WatchSupportsAutoPlaylistPlayback",
    "WatchSupportsHighQualityClockFaceGraphics",
    "WatchSupportsListeningOnGesture",
    "WatchSupportsMusicStreaming",
    "WatchSupportsSiriCommute",
    "WiFiCallingCapability",
    "WiFiCapability",
    "WifiAddress",
    "WifiAddressData",
    "WifiAntennaSKUVersion",
    "WifiCallingSecondaryDeviceCapability",
    "WifiChipset",
    "WifiFirmwareVersion",
    "WifiVendor",
    "WirelessBoardSnum",
    "WirelessChargingCapability",
    "YonkersChipID",
    "YonkersSerialNumber",
    "YonkersUID",
    "YouTubeCapability",
    "YouTubePluginCapability",
    "accelerometer",
    "accessibility",
    "additional-text-tones",
    "aggregate-cam-photo-zoom",
    "aggregate-cam-video-zoom",
    "airDropRestriction",
    "airplay-mirroring",
    "airplay-no-mirroring",
    "all-features",
    "allow-32bit-apps",
    "ambient-light-sensor",
    "ane",
    "any-telephony",
    "apn",
    "apple-internal-install",
    "applicationInstallation",
    "arkit",
    "arm64",
    "armv6",
    "armv7",
    "armv7s",
    "assistant",
    "auto-focus",
    "auto-focus-camera",
    "baseband-chipset",
    "bitrate-2g",
    "bitrate-3g",
    "bitrate-lte",
    "bitrate-wifi",
    "bluetooth",
    "bluetooth-le",
    "board-id",
    "boot-manifest-hash",
    "boot-nonce",
    "builtin-mics",
    "c2k-device",
    "calibration",
    "call-forwarding",
    "call-waiting",
    "caller-id",
    "camera-flash",
    "camera-front",
    "camera-front-flash",
    "camera-rear",
    "cameraRestriction",
    "car-integration",
    "cell-broadcast",
    "cellular-data",
    "certificate-production-status",
    "certificate-security-mode",
    "chip-id",
    "class",
    "closed-loop",
    "config-number",
    "contains-cellular-radio",
    "crypto-hash-method",
    "dali-mode",
    "data-plan",
    "debug-board-revision",
    "delay-sleep-for-headset-click",
    "device-color-policy",
    "device-colors",
    "device-name",
    "device-name-localized",
    "dictation",
    "die-id",
    "display-mirroring",
    "display-rotation",
    "displayport",
    "does-not-support-gamekit",
    "effective-production-status",
    "effective-production-status-ap",
    "effective-production-status-sep",
    "effective-security-mode",
    "effective-security-mode-ap",
    "effective-security-mode-sep",
    "enc-top-type",
    "encode-aac",
    "encrypted-data-partition",
    "enforce-googlemail",
    "enforce-shutter-click",
    "euicc-chip-id",
    "explicitContentRestriction",
    "face-detection-support",
    "fast-switch-options",
    "fcc-logos-via-software",
    "fcm-type",
    "firmware-version",
    "flash",
    "front-auto-hdr",
    "front-burst",
    "front-burst-image-duration",
    "front-facing-camera",
    "front-flash-capability",
    "front-hdr",
    "front-hdr-on",
    "front-max-video-fps-1080p",
    "front-max-video-fps-4k",
    "front-max-video-fps-720p",
    "front-max-video-zoom",
    "front-slowmo",
    "full-6",
    "function-button_halleffect",
    "function-button_ringerab",
    "gamekit",
    "gas-gauge-battery",
    "gps",
    "gps-capable",
    "green-tea",
    "gyroscope",
    "h264-encoder",
    "hall-effect-sensor",
    "haptics",
    "hardware-keyboard",
    "has-sphere",
    "hd-video-capture",
    "hdr-image-capture",
    "healthkit",
    "hearingaid-audio-equalization",
    "hearingaid-low-energy-audio",
    "hearingaid-power-reduction",
    "hiccough-interval",
    "hide-non-default-apps",
    "hidpi",
    "home-button-type",
    "homescreen-wallpaper",
    "hw-encode-snapshots",
    "hw-snapshots-need-purplegfx",
    "iAP2Capability",
    "iPadCapability",
    "iTunesFamilyID",
    "iap2-protocol-supported",
    "image4-supported",
    "international-settings",
    "io-surface-backed-images",
    "ipad",
    "kConferenceCallType",
    "kSimultaneousCallAndDataCurrentlySupported",
    "kSimultaneousCallAndDataSupported",
    "large-format-phone",
    "live-effects",
    "live-photo-capture",
    "load-thumbnails-while-scrolling",
    "location-reminders",
    "location-services",
    "low-power-wallet-mode",
    "lte-device",
    "magnetometer",
    "main-screen-class",
    "main-screen-height",
    "main-screen-orientation",
    "main-screen-pitch",
    "main-screen-scale",
    "main-screen-width",
    "marketing-name",
    "mesa",
    "metal",
    "microphone",
    "mix-n-match-prevention-status",
    "mms",
    "modelIdentifier",
    "multi-touch",
    "multitasking",
    "multitasking-gestures",
    "n78a-mode",
    "name",
    "navigation",
    "nfc",
    "nfcWithRadio",
    "nike-ipod",
    "nike-support",
    "no-coreroutine",
    "no-hi-res-buildings",
    "no-simplistic-road-mesh",
    "not-green-tea",
    "offline-dictation",
    "opal",
    "opengles-1",
    "opengles-2",
    "opengles-3",
    "opposed-power-vol-buttons",
    "ota-activation",
    "panorama",
    "peek-ui-width",
    "peer-peer",
    "personal-hotspot",
    "photo-adjustments",
    "photo-stream",
    "piezo-clicker",
    "pipelined-stillimage-capability",
    "platinum",
    "post-effects",
    "pressure",
    "prox-sensor",
    "proximity-sensor",
    "ptp-large-files",
    "public-key-accelerator",
    "rear-auto-hdr",
    "rear-burst",
    "rear-burst-image-duration",
    "rear-cam-telephoto-capability",
    "rear-facing-camera",
    "rear-hdr",
    "rear-hdr-on",
    "rear-max-slomo-video-fps-1080p",
    "rear-max-slomo-video-fps-720p",
    "rear-max-video-fps-1080p",
    "rear-max-video-fps-4k",
    "rear-max-video-fps-720p",
    "rear-max-video-frame_rate",
    "rear-max-video-zoom",
    "rear-slowmo",
    "regulatory-model-number",
    "ringer-switch",
    "role",
    's8000")',
    's8003")',
    "sandman-support",
    "screen-dimensions",
    "sensitive-ui",
    "shoebox",
    "sika-support",
    "sim",
    "sim-phonebook",
    "siri-gesture",
    "slow-letterpress-rendering",
    "sms",
    "software-bundle-version",
    "software-dimming-alpha",
    "stand-alone-contacts",
    "still-camera",
    "stockholm",
    "supports-always-listening",
    't7000")',
    "telephony",
    "telephony-maximum-generation",
    "thin-bezel",
    "tnr-mode-back",
    "tnr-mode-front",
    "touch-id",
    "tv-out-crossfade",
    "tv-out-settings",
    "ui-background-quality",
    "ui-no-parallax",
    "ui-no-procedural-wallpaper",
    "ui-pip",
    "ui-reachability",
    "ui-traffic-cheap-shaders",
    "ui-weather-quality",
    "umts-device",
    "unified-ipod",
    "unique-chip-id",
    "venice",
    "video-camera",
    "video-cap",
    "video-stills",
    "voice-control",
    "voip",
    "volume-buttons",
    "wapi",
    "watch-companion",
    "wi-fi",
    "wifi",
    "wifi-antenna-sku-info",
    "wifi-chipset",
    "wifi-module-sn",
    "wlan",
    "wlan.background-scan-cache",
    "youtube",
    "youtubePlugin",
]


class DiagnosticsService(LockdownService):
    """
    Provides an API to:
    * Query MobileGestalt & IORegistry keys.
    * Reboot, shutdown or put the device in sleep mode.
    """

    RSD_SERVICE_NAME = "com.apple.mobile.diagnostics_relay.shim.remote"
    SERVICE_NAME = "com.apple.mobile.diagnostics_relay"
    OLD_SERVICE_NAME = "com.apple.iosdiagnostics.relay"

    def __init__(self, lockdown: LockdownServiceProvider):
        service_name = self.SERVICE_NAME if isinstance(lockdown, LockdownClient) else self.RSD_SERVICE_NAME
        super().__init__(lockdown, service_name, service=None)

    async def _send_recv(self, request: dict) -> dict:
        return await self.service.send_recv_plist(request)

    async def mobilegestalt(self, keys: Optional[list[str]] = None) -> dict:
        if keys is None or len(keys) == 0:
            keys = MobileGestaltKeys
        response = await self._send_recv({"Request": "MobileGestalt", "MobileGestaltKeys": keys})

        if response["Diagnostics"]["MobileGestalt"]["Status"] == "MobileGestaltDeprecated":
            err_msg = f"failed to query MobileGestalt, MobileGestalt deprecated (iOS >= 17.4). Got response {response}"
            raise DeprecationError(err_msg)

        if (response["Status"] != "Success") or (response["Diagnostics"]["MobileGestalt"]["Status"] != "Success"):
            raise PyMobileDevice3Exception("failed to query MobileGestalt")

        response["Diagnostics"]["MobileGestalt"].pop("Status")

        return response["Diagnostics"]["MobileGestalt"]

    async def action(self, action: str) -> Optional[dict]:
        response = await self._send_recv({"Request": action})
        if response["Status"] != "Success":
            raise PyMobileDevice3Exception(f"failed to perform action: {action}")
        return response.get("Diagnostics")

    async def restart(self):
        await self.action("Restart")

    async def shutdown(self):
        await self.action("Shutdown")

    async def sleep(self):
        await self.action("Sleep")

    async def info(self, diag_type: str = "All") -> dict:
        return await self.action(diag_type)

    async def ioregistry(self, plane: Optional[str] = None, name: Optional[str] = None, ioclass: Optional[str] = None):
        d = {}

        if plane:
            d["CurrentPlane"] = plane

        if name:
            d["EntryName"] = name

        if ioclass:
            d["EntryClass"] = ioclass

        d["Request"] = "IORegistry"

        response = await self._send_recv(d)
        if response.get("Status") != "Success":
            raise PyMobileDevice3Exception(f"got invalid response: {response}")

        dd = response.get("Diagnostics")
        if dd:
            return dd.get("IORegistry")
        return None

    async def get_battery(self) -> dict:
        return await self.ioregistry(ioclass="IOPMPowerSource")

    async def get_wifi(self) -> dict:
        result = await self.ioregistry(name="AppleBCMWLANSkywalkInterface")
        if result:
            return result
        return await self.ioregistry(ioclass="IO80211Interface")



import asyncio
from io import BytesIO
from zipfile import ZIP_DEFLATED, ZipFile

import pytest

from pymobiledevice3.exceptions import AppInstallError
from pymobiledevice3.lockdown import LockdownClient, create_using_usbmux
from pymobiledevice3.services.afc import AfcService
from pymobiledevice3.services.installation_proxy import InstallationProxyService


def _make_minimal_ipa_bytes() -> bytes:
    buffer = BytesIO()
    with ZipFile(buffer, "w", ZIP_DEFLATED) as zip_file:
        zip_file.writestr("Payload/Test.app/Info.plist", b"")
    return buffer.getvalue()


@pytest.mark.asyncio
async def test_get_apps(lockdown: LockdownClient) -> None:
    async with InstallationProxyService(lockdown=lockdown) as installation_proxy:
        apps = await installation_proxy.get_apps()
        assert len(apps) > 1


@pytest.mark.asyncio
async def test_get_system_apps(lockdown: LockdownClient) -> None:
    async with InstallationProxyService(lockdown=lockdown) as installation_proxy:
        system_apps = await installation_proxy.get_apps(application_type="System")
        app_types = {app["ApplicationType"] for app in system_apps.values()}
        assert len(app_types) == 1
        assert app_types.pop() == "System"


@pytest.mark.asyncio
async def test_parallel_installations_cleanup(lockdown: LockdownClient, monkeypatch: pytest.MonkeyPatch) -> None:
    ipa_bytes = _make_minimal_ipa_bytes()
    captured_plists: list[dict] = []

    def wrap_send_plist(service, captured: list[dict]):
        original_send_plist = service.send_plist

        async def wrapped_send_plist(payload: dict) -> None:
            if payload.get("Command") == "Install":
                captured.append(payload.copy())
            await original_send_plist(payload)

        return wrapped_send_plist

    async with (
        await create_using_usbmux(serial=lockdown.identifier) as inner_lockdown,
        InstallationProxyService(lockdown=inner_lockdown) as first,
        InstallationProxyService(lockdown=inner_lockdown) as second,
    ):
        monkeypatch.setattr(first.service, "send_plist", wrap_send_plist(first.service, captured_plists))
        monkeypatch.setattr(second.service, "send_plist", wrap_send_plist(second.service, captured_plists))

        results = await asyncio.gather(
            first.install_from_bytes(ipa_bytes),
            second.install_from_bytes(ipa_bytes),
            return_exceptions=True,
        )

    # https://github.com/doronz88/pymobiledevice3/issues/1595
    # will fails with:
    # >       assert all(isinstance(result, AppInstallError) for result in results), f"Expected both installations to fail with AppInstallError, got {results}"
    # E       AssertionError: Expected both installations to fail with AppInstallError, got [AppInstallError('APIInternalError: Error Domain=IXErrorDomain Code=13 "Failed to get bundle ID from /private/var/tmp/com.apple.appinstall.temp/temp.oX9BMG/Extracted/Payload/Test.app" UserInfo={NSLocalizedDescription=Failed to get bundle ID from /private/var/tmp/com.apple.appinstall.temp/temp.oX9BMG/Extracted/Payload/Test.app, FunctionName=+[IXPlaceholder _placeholderForBundle:client:withParent:installType:metadata:placeholderType:mayBeDeltaPackage:isFromSerializedPlaceholder:location:error:], SourceFileLine=752, NSLocalizedFailureReason=Missing bundle ID.}'), RuntimeError('readexactly() called while another coroutine is already waiting for incoming data')]
    # E       assert False
    # E        +  where False = all(<generator object test_parallel_installations_cleanup.<locals>.<genexpr> at 0x1121d7760>)
    assert all(isinstance(result, AppInstallError) for result in results), (
        f"Expected both installations to fail with AppInstallError, got {results}"
    )
    assert len(captured_plists) == 2, f"Expected 2 captured plists, got {len(captured_plists)}"

    package_paths = [payload["PackagePath"] for payload in captured_plists]
    assert len(set(package_paths)) == 2, f"Expected different package paths for each installation, got {package_paths}"

    async with AfcService(lockdown=lockdown) as afc:
        for package_path in package_paths:
            assert not await afc.exists(package_path), f"Expected package path {package_path} to be cleaned up"



#!/usr/bin/env python3
import logging
import time
from typing import Optional

from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.lockdown_service_provider import LockdownServiceProvider


class HeartbeatService:
    """
    Use to keep an active connection with lockdownd
    """

    SERVICE_NAME = "com.apple.mobile.heartbeat"
    RSD_SERVICE_NAME = "com.apple.mobile.heartbeat.shim.remote"

    def __init__(self, lockdown: LockdownServiceProvider):
        self.logger = logging.getLogger(__name__)
        self.lockdown = lockdown

        if isinstance(lockdown, LockdownClient):
            self.service_name = self.SERVICE_NAME
        else:
            self.service_name = self.RSD_SERVICE_NAME

    async def start(self, interval: Optional[float] = None) -> None:
        start = time.time()
        service = await self.lockdown.start_lockdown_service(self.service_name)

        while True:
            response = await service.recv_plist()
            self.logger.debug(response)

            if interval is not None and time.time() >= start + interval:
                break

            await service.send_plist({"Command": "Polo"})



import asyncio
from contextlib import suppress
from pathlib import Path
from typing import Annotated, Optional

import typer
from defusedxml import ElementTree as DefusedET
from typer_injector import Depends, InjectingTyper

from pymobiledevice3 import usbmux
from pymobiledevice3.cli.cli_common import ServiceProviderDep, async_command, print_json
from pymobiledevice3.exceptions import ConnectionFailedError
from pymobiledevice3.lockdown_service_provider import LockdownServiceProvider
from pymobiledevice3.services.dvt.testmanaged.xcuitest import TestConfig, XCUITestService
from pymobiledevice3.services.wda import DEFAULT_WDA_PORT, WdaServiceClient
from pymobiledevice3.utils import get_asyncio_loop

cli = InjectingTyper(
    name="wda",
    help="Interact with WebDriverAgent (launch apps and tap elements).",
    no_args_is_help=True,
)


def wda_client_dep(
    service_provider: ServiceProviderDep,
    port: Annotated[
        int,
        typer.Option(
            "--port",
            help="WDA device port to connect to.",
        ),
    ] = DEFAULT_WDA_PORT,
    timeout: Annotated[
        float,
        typer.Option(help="HTTP timeout in seconds."),
    ] = 10.0,
) -> WdaServiceClient:
    return WdaServiceClient(service_provider=service_provider, port=port, timeout=timeout)


WdaClientDep = Annotated[
    WdaServiceClient,
    Depends(wda_client_dep),
]


async def wait_for_xctest_app(service_provider: LockdownServiceProvider, xctrunner: str) -> asyncio.Future:
    cfg = await TestConfig.create_for(service_provider, runner_bundle_id=xctrunner)

    ts = XCUITestService(service_provider)
    xctrunner_task = asyncio.create_task(ts.run(cfg), name="wda-xctrunner")
    startup_timeout = 30.0
    deadline = asyncio.get_running_loop().time() + startup_timeout

    try:
        while True:
            # If XCTest setup fails, surface the failure instead of waiting forever.
            if xctrunner_task.done():
                xctrunner_task.result()
                raise RuntimeError("XCUITest runner exited before WDA became reachable")
            if asyncio.get_running_loop().time() >= deadline:
                raise TimeoutError(f"WDA did not become reachable on port {DEFAULT_WDA_PORT} within 30.0 seconds")

            device = await usbmux.select_device(service_provider.udid)
            try:
                await device.connect(DEFAULT_WDA_PORT)
            except ConnectionFailedError:
                await asyncio.sleep(0.1)
            else:
                break
    except Exception:
        xctrunner_task.cancel()
        with suppress(asyncio.CancelledError):
            await xctrunner_task
        raise

    return xctrunner_task


def wda_xctest_dep(
    service_provider: ServiceProviderDep,
    xctrunner: Annotated[
        Optional[str],
        typer.Option(
            "--xctrunner",
            "-xc",
            help="Bundle id of an XCUITest runner to start (e.g. com.facebook.WebDriverAgentRunner.xctrunner).",
        ),
    ] = None,
) -> Optional[asyncio.Future]:
    if xctrunner is None:
        return None
    return get_asyncio_loop().run_until_complete(wait_for_xctest_app(service_provider, xctrunner))


WdaXcRunnerDep = Annotated[
    Optional[asyncio.Future],
    Depends(wda_xctest_dep),
]


async def _cleanup_xctrunner(xctrunner: WdaXcRunnerDep) -> None:
    if xctrunner is None:
        return
    xctrunner.cancel()
    with suppress(asyncio.CancelledError):
        await xctrunner


@cli.command("launch")
@async_command
async def wda_launch(
    client: WdaClientDep,
    _xctrunner: WdaXcRunnerDep,
    bundle_id: str,
) -> None:
    """Launch an app by starting a WDA session and print the session id."""
    session_id = await client.start_session(bundle_id=bundle_id)
    print(session_id)
    await _cleanup_xctrunner(_xctrunner)


@cli.command("tap")
@async_command
async def wda_tap(
    client: WdaClientDep,
    _xctrunner: WdaXcRunnerDep,
    selector: str,
    using: Annotated[
        str,
        typer.Option(
            "--using",
            "-u",
            help="Element lookup strategy (e.g. 'accessibility id', 'name', 'label', 'xpath').",
        ),
    ] = "accessibility id",
    session_id: Annotated[
        Optional[str],
        typer.Option(
            "--session-id",
            "-s",
            help="Existing WDA session id (omit to create a new session).",
        ),
    ] = None,
    bundle_id: Annotated[
        Optional[str],
        typer.Option(
            "--bundle-id",
            help="Bundle id to launch if a new session is created.",
        ),
    ] = None,
) -> None:
    """Tap an element (typically a button) using a WDA selector."""
    if session_id is None:
        session_id = await client.start_session(bundle_id=bundle_id)
    else:
        client.session_id = session_id
    element_id = await client.find_element(using=using, value=selector, session_id=session_id)
    await client.click(element_id=element_id, session_id=session_id)
    await _cleanup_xctrunner(_xctrunner)


@cli.command("press")
@async_command
async def wda_press(
    client: WdaClientDep,
    _xctrunner: WdaXcRunnerDep,
    names: Annotated[
        list[str],
        typer.Argument(
            help="One or more device buttons to press in order (e.g. home lock volumeup).",
        ),
    ],
    session_id: Annotated[
        Optional[str],
        typer.Option(
            "--session-id",
            "-s",
            help="Existing WDA session id (optional; some servers require it).",
        ),
    ] = None,
) -> None:
    """Press one or more device buttons (e.g. home, volumeup, volumedown, lock)."""
    if session_id is None:
        session_id = await client.start_session()
    else:
        client.session_id = session_id
    for name in names:
        await client.press_button(name=name, session_id=session_id)
    await _cleanup_xctrunner(_xctrunner)


@cli.command("unlock")
@async_command
async def wda_unlock(
    client: WdaClientDep,
    _xctrunner: WdaXcRunnerDep,
    session_id: Annotated[
        Optional[str],
        typer.Option(
            "--session-id",
            "-s",
            help="Existing WDA session id (optional; used for session-scoped unlock endpoints).",
        ),
    ] = None,
) -> None:
    """Unlock the device."""
    await client.unlock(session_id=session_id)
    await _cleanup_xctrunner(_xctrunner)


@cli.command("list-items")
@async_command
async def wda_list_items(
    client: WdaClientDep,
    _xctrunner: WdaXcRunnerDep,
    session_id: Annotated[
        Optional[str],
        typer.Option(
            "--session-id",
            "-s",
            help="Existing WDA session id (optional).",
        ),
    ] = None,
    hittable_only: Annotated[
        bool,
        typer.Option(
            "--hittable-only",
            help="Show only elements marked hittable by WDA.",
        ),
    ] = False,
    clickable: Annotated[
        bool,
        typer.Option(
            "--clickable",
            help="Filter by a clickable type allowlist + visible/enabled.",
        ),
    ] = False,
    types: Annotated[
        Optional[list[str]],
        typer.Option(
            "--types",
            help="Comma-separated element type allowlist (e.g. Icon,Button,Cell).",
        ),
    ] = None,
    with_rect: Annotated[
        bool,
        typer.Option(
            "--with-rect",
            help="Include element bounds (x,y,width,height) if present.",
        ),
    ] = False,
) -> None:
    """Show tappable WDA elements (use with 'tap' selectors)."""
    source = await client.get_source(session_id=session_id)
    root = DefusedET.fromstring(source)

    clickable_types = {
        "XCUIElementTypeButton",
        "XCUIElementTypeIcon",
        "XCUIElementTypeCell",
        "XCUIElementTypeLink",
        "XCUIElementTypeSwitch",
        "XCUIElementTypeTextField",
        "XCUIElementTypeSecureTextField",
        "XCUIElementTypeImage",
    }
    if types:
        custom = set()
        for entry in types:
            for item in entry.split(","):
                item = item.strip()
                if not item:
                    continue
                if item.startswith("XCUIElementType"):
                    custom.add(item)
                else:
                    custom.add(f"XCUIElementType{item}")
        if custom:
            clickable_types = custom

    items = []
    for elem in root.iter():
        attrs = elem.attrib
        elem_type = elem.tag
        enabled = attrs.get("enabled")
        visible = attrs.get("visible")
        hittable = attrs.get("hittable")
        if hittable_only and hittable != "true":
            continue
        if clickable:
            if visible != "true" or enabled != "true":
                continue
            if elem_type not in clickable_types:
                continue
        name = attrs.get("name")
        label = attrs.get("label")
        value = attrs.get("value")
        rect = None
        if with_rect:
            rect = {
                "x": attrs.get("x"),
                "y": attrs.get("y"),
                "width": attrs.get("width"),
                "height": attrs.get("height"),
            }
            if all(v is None for v in rect.values()):
                rect = None
        if not (name or label or value or rect):
            continue
        item = {
            "type": elem_type,
            "name": name,
            "label": label,
            "value": value,
            "enabled": enabled,
            "visible": visible,
            "hittable": hittable,
        }
        if with_rect:
            item["rect"] = rect
        items.append(item)

    print_json(items)
    await _cleanup_xctrunner(_xctrunner)


@cli.command("screenshot")
@async_command
async def wda_screenshot(
    client: WdaClientDep,
    _xctrunner: WdaXcRunnerDep,
    out: Path,
    session_id: Annotated[
        Optional[str],
        typer.Option(
            "--session-id",
            "-s",
            help="Existing WDA session id (optional).",
        ),
    ] = None,
) -> None:
    """Save a screenshot via WDA."""
    data = await client.get_screenshot(session_id=session_id)
    out.write_bytes(data)
    await _cleanup_xctrunner(_xctrunner)


@cli.command("status")
@async_command
async def wda_status(
    client: WdaClientDep,
    _xctrunner: WdaXcRunnerDep,
) -> None:
    """Show WDA status."""
    print_json(await client.get_status())
    await _cleanup_xctrunner(_xctrunner)


@cli.command("type")
@async_command
async def wda_type(
    client: WdaClientDep,
    _xctrunner: WdaXcRunnerDep,
    text: str,
    session_id: Annotated[
        Optional[str],
        typer.Option(
            "--session-id",
            "-s",
            help="Existing WDA session id (omit to create a new session).",
        ),
    ] = None,
    bundle_id: Annotated[
        Optional[str],
        typer.Option(
            "--bundle-id",
            help="Bundle id to launch if a new session is created.",
        ),
    ] = None,
) -> None:
    """Type text into the focused element."""
    if session_id is None:
        session_id = await client.start_session(bundle_id=bundle_id)
    else:
        client.session_id = session_id
    await client.send_keys(text, session_id=session_id)
    await _cleanup_xctrunner(_xctrunner)


@cli.command("swipe")
@async_command
async def wda_swipe(
    client: WdaClientDep,
    _xctrunner: WdaXcRunnerDep,
    start_x: int,
    start_y: int,
    end_x: int,
    end_y: int,
    duration: Annotated[
        float,
        typer.Option(
            "--duration",
            "-d",
            help="Swipe duration in seconds.",
        ),
    ] = 0.2,
    session_id: Annotated[
        Optional[str],
        typer.Option(
            "--session-id",
            "-s",
            help="Existing WDA session id (omit to create a new session).",
        ),
    ] = None,
    bundle_id: Annotated[
        Optional[str],
        typer.Option(
            "--bundle-id",
            help="Bundle id to launch if a new session is created.",
        ),
    ] = None,
) -> None:
    """Swipe from one coordinate to another."""
    if session_id is None:
        session_id = await client.start_session(bundle_id=bundle_id)
    else:
        client.session_id = session_id
    await client.swipe(start_x, start_y, end_x, end_y, duration=duration, session_id=session_id)
    await _cleanup_xctrunner(_xctrunner)


@cli.command("window-size")
@async_command
async def wda_window_size(
    client: WdaClientDep,
    _xctrunner: WdaXcRunnerDep,
    session_id: Annotated[
        Optional[str],
        typer.Option(
            "--session-id",
            "-s",
            help="Existing WDA session id (omit to create a new session).",
        ),
    ] = None,
    bundle_id: Annotated[
        Optional[str],
        typer.Option(
            "--bundle-id",
            help="Bundle id to launch if a new session is created.",
        ),
    ] = None,
) -> None:
    """Show window size (screen dimensions)."""
    if session_id is None:
        session_id = await client.start_session(bundle_id=bundle_id)
    else:
        client.session_id = session_id
    print_json(await client.get_window_size(session_id=session_id))
    await _cleanup_xctrunner(_xctrunner)



import asyncio

import pytest

from pymobiledevice3.lockdown import SERVICE_PORT, LockdownClient
from pymobiledevice3.service_connection import ServiceConnection
from pymobiledevice3.tcp_forwarder import TcpForwarderBase, UsbmuxTcpForwarder

FREE_PORT = 3582


class DummyForwarder(TcpForwarderBase):
    async def _establish_remote_connection(self) -> ServiceConnection:
        raise NotImplementedError


async def attempt_local_connection(port: int) -> None:
    _reader, writer = await asyncio.open_connection("127.0.0.1", port)
    writer.close()
    await writer.wait_closed()


@pytest.mark.parametrize("dst_port", [FREE_PORT, SERVICE_PORT])
@pytest.mark.asyncio
async def test_tcp_forwarder_bad_port(lockdown: LockdownClient, dst_port: int) -> None:
    # start forwarder
    forwarder = UsbmuxTcpForwarder(lockdown.udid, dst_port, FREE_PORT)
    task = asyncio.create_task(forwarder.start())

    try:
        # wait for it to actually start listening
        for _ in range(100):
            if forwarder.server is not None:
                break
            await asyncio.sleep(0.01)
        assert forwarder.server is not None
        await attempt_local_connection(FREE_PORT)

    finally:
        # tell it to stop
        forwarder.stop()
        # make sure it stops
        await asyncio.wait_for(task, timeout=5)


@pytest.mark.asyncio
async def test_tcp_forwarder_ephemeral_port_and_asyncio_event() -> None:
    listening_event = asyncio.Event()
    forwarder = DummyForwarder(0, listening_event=listening_event)
    task = asyncio.create_task(forwarder.start())

    try:
        await asyncio.wait_for(listening_event.wait(), timeout=5)
        assert forwarder.server is not None
        assert forwarder.listening_port is not None
        assert forwarder.listening_port > 0
    finally:
        forwarder.stop()
        await asyncio.wait_for(task, timeout=5)



import asyncio
import traceback
from functools import wraps
from pathlib import Path
from typing import Any, Callable, Optional

import IPython
import requests
from construct import Int8ul, Int16ul, Int32ul, Int64ul, Select
from tqdm import tqdm
from traitlets.config import Config


def plist_access_path(d, path: tuple, type_=None, required=False):
    for component in path:
        d = d.get(component)
        if d is None:
            break

    if type_ is bool and isinstance(d, str):
        if d.lower() not in ("true", "false"):
            raise ValueError()
        d = d.lower() == "true"
    elif type_ is not None and not isinstance(d, type_):
        # wrong type
        d = None

    if d is None and required:
        raise KeyError(f"path: {path} doesn't exist in given plist object")

    return d


def bytes_to_uint(b: bytes):
    return Select(u64=Int64ul, u32=Int32ul, u16=Int16ul, u8=Int8ul).parse(b)


def try_decode(s: bytes):
    try:
        return s.decode("utf8")
    except UnicodeDecodeError:
        return s


def asyncio_print_traceback(f: Callable):
    @wraps(f)
    async def wrapper(*args, **kwargs):
        try:
            return await f(*args, **kwargs)
        except (Exception, RuntimeError) as e:
            if not isinstance(e, asyncio.CancelledError):
                traceback.print_exc()
            raise

    return wrapper


_ASYNCIO_LOOP: Optional[asyncio.AbstractEventLoop] = None


def get_asyncio_loop() -> asyncio.AbstractEventLoop:
    global _ASYNCIO_LOOP
    if _ASYNCIO_LOOP is None or _ASYNCIO_LOOP.is_closed():
        _ASYNCIO_LOOP = asyncio.new_event_loop()
    return _ASYNCIO_LOOP


def run_in_loop(coro):
    return get_asyncio_loop().run_until_complete(coro)


def start_ipython_shell(*, user_ns: Optional[dict[str, Any]] = None, header: Optional[str] = None) -> None:
    # Keep IPython autoawait on the same loop used by CLI async wrappers.
    config = Config()
    config.InteractiveShell.loop_runner = run_in_loop
    if header is not None:
        print(header)
    IPython.start_ipython(argv=[], config=config, user_ns=user_ns or {})


def file_download(url: str, outfile: Path, chunk_size=1024) -> None:
    resp = requests.get(url, stream=True)
    total = int(resp.headers.get("content-length", 0))
    with (
        outfile.open("wb") as file,
        tqdm(
            desc=outfile.name,
            total=total,
            unit="iB",
            unit_scale=True,
            unit_divisor=1024,
        ) as bar,
    ):
        for data in resp.iter_content(chunk_size=chunk_size):
            size = file.write(data)
            bar.update(size)



import subprocess
import sys

import pytest
from typer.testing import CliRunner

from pymobiledevice3 import __main__

pytestmark = [pytest.mark.cli]


@pytest.mark.xfail(reason="Looks like click broke something")
def test_cli_main_interface():
    runner = CliRunner()
    r1 = runner.invoke(__main__.app, ["--help"])
    assert r1.exit_code == 0

    r2 = runner.invoke(__main__.app)
    assert r2.exit_code == 0


@pytest.mark.parametrize(
    "keyword,suggestions",
    [
        ("kill", ["developer dvt kill", "developer dvt pkill"]),
        ("sysdi", ["crash sysdiagnose"]),
        (
            "shell",
            [
                "afc shell",
                "crash shell",
                "developer accessibility shell",
                "developer dvt shell",
                "developer shell",
                "restore shell",
                "springboard shell",
                "webinspector js-shell",
                "webinspector shell",
            ],
        ),
        (
            "shall",
            [
                "afc shell",
                "crash pull",
                "crash shell",
                "apps install",
                "crash ls",
                "afc pull",
                "restore shell",
                "apps uninstall",
                "profile install",
                "developer shell",
            ],
        ),
    ],
)
def test_cli_suggestions(keyword, suggestions):
    output = subprocess.run(
        [sys.executable, "-m", "pymobiledevice3", keyword],
        capture_output=True,
        text=True,
    )
    for suggestion in suggestions:
        assert suggestion in output.stderr


@pytest.mark.parametrize("group", __main__.CLI_GROUPS.keys())
def test_cli_groups(group):
    runner = CliRunner()
    group_help_result = runner.invoke(__main__.app, [group, "--help"])
    assert group_help_result.exit_code == 0


@pytest.mark.parametrize("group", __main__.CLI_GROUPS.keys())
def test_cli_from_python_m_flag(group):
    subprocess.run([sys.executable, "-m", "pymobiledevice3", group, "--help"], check=True)



import sys
from enum import Enum


class ConnectionType(str, Enum):
    USB = "usb"
    WIFI = "wifi"


class TunnelProtocol(str, Enum):
    TCP = "tcp"
    QUIC = "quic"

    # TODO: make only TCP the default once 3.12 becomes deprecated
    DEFAULT = TCP if sys.version_info >= (3, 13) else QUIC



from collections import namedtuple

IRecvDevice = namedtuple("IRecvDevice", "product_type hardware_model board_id chip_id display_name")

IRECV_DEVICES = (
    # iPhone
    IRecvDevice(
        product_type="iPhone1,1", hardware_model="m68ap", board_id=0x00, chip_id=0x8900, display_name="iPhone 2G"
    ),
    IRecvDevice(
        product_type="iPhone1,2", hardware_model="n82ap", board_id=0x04, chip_id=0x8900, display_name="iPhone 3G"
    ),
    IRecvDevice(
        product_type="iPhone2,1", hardware_model="n88ap", board_id=0x00, chip_id=0x8920, display_name="iPhone 3Gs"
    ),
    IRecvDevice(
        product_type="iPhone3,1", hardware_model="n90ap", board_id=0x00, chip_id=0x8930, display_name="iPhone 4 (GSM)"
    ),
    IRecvDevice(
        product_type="iPhone3,2",
        hardware_model="n90bap",
        board_id=0x04,
        chip_id=0x8930,
        display_name="iPhone 4 (GSM) R2 2012",
    ),
    IRecvDevice(
        product_type="iPhone3,3", hardware_model="n92ap", board_id=0x06, chip_id=0x8930, display_name="iPhone 4 (CDMA)"
    ),
    IRecvDevice(
        product_type="iPhone4,1", hardware_model="n94ap", board_id=0x08, chip_id=0x8940, display_name="iPhone 4s"
    ),
    IRecvDevice(
        product_type="iPhone5,1", hardware_model="n41ap", board_id=0x00, chip_id=0x8950, display_name="iPhone 5 (GSM)"
    ),
    IRecvDevice(
        product_type="iPhone5,2",
        hardware_model="n42ap",
        board_id=0x02,
        chip_id=0x8950,
        display_name="iPhone 5 (Global)",
    ),
    IRecvDevice(
        product_type="iPhone5,3", hardware_model="n48ap", board_id=0x0A, chip_id=0x8950, display_name="iPhone 5c (GSM)"
    ),
    IRecvDevice(
        product_type="iPhone5,4",
        hardware_model="n49ap",
        board_id=0x0E,
        chip_id=0x8950,
        display_name="iPhone 5c (Global)",
    ),
    IRecvDevice(
        product_type="iPhone6,1", hardware_model="n51ap", board_id=0x00, chip_id=0x8960, display_name="iPhone 5s (GSM)"
    ),
    IRecvDevice(
        product_type="iPhone6,2",
        hardware_model="n53ap",
        board_id=0x02,
        chip_id=0x8960,
        display_name="iPhone 5s (Global)",
    ),
    IRecvDevice(
        product_type="iPhone7,1", hardware_model="n56ap", board_id=0x04, chip_id=0x7000, display_name="iPhone 6 Plus"
    ),
    IRecvDevice(
        product_type="iPhone7,2", hardware_model="n61ap", board_id=0x06, chip_id=0x7000, display_name="iPhone 6"
    ),
    IRecvDevice(
        product_type="iPhone8,1", hardware_model="n71ap", board_id=0x04, chip_id=0x8000, display_name="iPhone 6s"
    ),
    IRecvDevice(
        product_type="iPhone8,1", hardware_model="n71map", board_id=0x04, chip_id=0x8003, display_name="iPhone 6s"
    ),
    IRecvDevice(
        product_type="iPhone8,2", hardware_model="n66ap", board_id=0x06, chip_id=0x8000, display_name="iPhone 6s Plus"
    ),
    IRecvDevice(
        product_type="iPhone8,2", hardware_model="n66map", board_id=0x06, chip_id=0x8003, display_name="iPhone 6s Plus"
    ),
    IRecvDevice(
        product_type="iPhone8,4",
        hardware_model="n69ap",
        board_id=0x02,
        chip_id=0x8003,
        display_name="iPhone SE (1st gen)",
    ),
    IRecvDevice(
        product_type="iPhone8,4",
        hardware_model="n69uap",
        board_id=0x02,
        chip_id=0x8000,
        display_name="iPhone SE (1st gen)",
    ),
    IRecvDevice(
        product_type="iPhone9,1",
        hardware_model="d10ap",
        board_id=0x08,
        chip_id=0x8010,
        display_name="iPhone 7 (Global)",
    ),
    IRecvDevice(
        product_type="iPhone9,2",
        hardware_model="d11ap",
        board_id=0x0A,
        chip_id=0x8010,
        display_name="iPhone 7 Plus (Global)",
    ),
    IRecvDevice(
        product_type="iPhone9,3", hardware_model="d101ap", board_id=0x0C, chip_id=0x8010, display_name="iPhone 7 (GSM)"
    ),
    IRecvDevice(
        product_type="iPhone9,4",
        hardware_model="d111ap",
        board_id=0x0E,
        chip_id=0x8010,
        display_name="iPhone 7 Plus (GSM)",
    ),
    IRecvDevice(
        product_type="iPhone10,1",
        hardware_model="d20ap",
        board_id=0x02,
        chip_id=0x8015,
        display_name="iPhone 8 (Global)",
    ),
    IRecvDevice(
        product_type="iPhone10,2",
        hardware_model="d21ap",
        board_id=0x04,
        chip_id=0x8015,
        display_name="iPhone 8 Plus (Global)",
    ),
    IRecvDevice(
        product_type="iPhone10,3",
        hardware_model="d22ap",
        board_id=0x06,
        chip_id=0x8015,
        display_name="iPhone X (Global)",
    ),
    IRecvDevice(
        product_type="iPhone10,4", hardware_model="d201ap", board_id=0x0A, chip_id=0x8015, display_name="iPhone 8 (GSM)"
    ),
    IRecvDevice(
        product_type="iPhone10,5",
        hardware_model="d211ap",
        board_id=0x0C,
        chip_id=0x8015,
        display_name="iPhone 8 Plus (GSM)",
    ),
    IRecvDevice(
        product_type="iPhone10,6", hardware_model="d221ap", board_id=0x0E, chip_id=0x8015, display_name="iPhone X (GSM)"
    ),
    IRecvDevice(
        product_type="iPhone11,2", hardware_model="d321ap", board_id=0x0E, chip_id=0x8020, display_name="iPhone XS"
    ),
    IRecvDevice(
        product_type="iPhone11,4",
        hardware_model="d331ap",
        board_id=0x0A,
        chip_id=0x8020,
        display_name="iPhone XS Max (China)",
    ),
    IRecvDevice(
        product_type="iPhone11,6", hardware_model="d331pap", board_id=0x1A, chip_id=0x8020, display_name="iPhone XS Max"
    ),
    IRecvDevice(
        product_type="iPhone11,8", hardware_model="n841ap", board_id=0x0C, chip_id=0x8020, display_name="iPhone XR"
    ),
    IRecvDevice(
        product_type="iPhone12,1", hardware_model="n104ap", board_id=0x04, chip_id=0x8030, display_name="iPhone 11"
    ),
    IRecvDevice(
        product_type="iPhone12,3", hardware_model="d421ap", board_id=0x06, chip_id=0x8030, display_name="iPhone 11 Pro"
    ),
    IRecvDevice(
        product_type="iPhone12,5",
        hardware_model="d431ap",
        board_id=0x02,
        chip_id=0x8030,
        display_name="iPhone 11 Pro Max",
    ),
    IRecvDevice(
        product_type="iPhone12,8",
        hardware_model="d79ap",
        board_id=0x10,
        chip_id=0x8030,
        display_name="iPhone SE (2nd gen)",
    ),
    IRecvDevice(
        product_type="iPhone13,1", hardware_model="d52gap", board_id=0x0A, chip_id=0x8101, display_name="iPhone 12 mini"
    ),
    IRecvDevice(
        product_type="iPhone13,2", hardware_model="d53gap", board_id=0x0C, chip_id=0x8101, display_name="iPhone 12"
    ),
    IRecvDevice(
        product_type="iPhone13,3", hardware_model="d53pap", board_id=0x0E, chip_id=0x8101, display_name="iPhone 12 Pro"
    ),
    IRecvDevice(
        product_type="iPhone13,4",
        hardware_model="d54pap",
        board_id=0x08,
        chip_id=0x8101,
        display_name="iPhone 12 Pro Max",
    ),
    IRecvDevice(
        product_type="iPhone14,2", hardware_model="d63ap", board_id=0x0C, chip_id=0x8110, display_name="iPhone 13 Pro"
    ),
    IRecvDevice(
        product_type="iPhone14,3",
        hardware_model="d64ap",
        board_id=0x0E,
        chip_id=0x8110,
        display_name="iPhone 13 Pro Max",
    ),
    IRecvDevice(
        product_type="iPhone14,4", hardware_model="d16ap", board_id=0x08, chip_id=0x8110, display_name="iPhone 13 mini"
    ),
    IRecvDevice(
        product_type="iPhone14,5", hardware_model="d17ap", board_id=0x0A, chip_id=0x8110, display_name="iPhone 13"
    ),
    IRecvDevice(
        product_type="iPhone14,6",
        hardware_model="d49ap",
        board_id=0x10,
        chip_id=0x8110,
        display_name="iPhone SE (3rd gen)",
    ),
    IRecvDevice(
        product_type="iPhone14,7", hardware_model="d27ap", board_id=0x18, chip_id=0x8110, display_name="iPhone 14"
    ),
    IRecvDevice(
        product_type="iPhone14,8", hardware_model="d28ap", board_id=0x1A, chip_id=0x8110, display_name="iPhone 14 Plus"
    ),
    IRecvDevice(
        product_type="iPhone15,2", hardware_model="d73ap", board_id=0x0C, chip_id=0x8120, display_name="iPhone 14 Pro"
    ),
    IRecvDevice(
        product_type="iPhone15,3",
        hardware_model="d74ap",
        board_id=0x0E,
        chip_id=0x8120,
        display_name="iPhone 14 Pro Max",
    ),
    IRecvDevice(
        product_type="iPhone15,4", hardware_model="d37ap", board_id=0x08, chip_id=0x8120, display_name="iPhone 15"
    ),
    IRecvDevice(
        product_type="iPhone15,5", hardware_model="d38ap", board_id=0x0A, chip_id=0x8120, display_name="iPhone 15 Plus"
    ),
    IRecvDevice(
        product_type="iPhone16,1", hardware_model="d83ap", board_id=0x04, chip_id=0x8130, display_name="iPhone 15 Pro"
    ),
    IRecvDevice(
        product_type="iPhone16,2",
        hardware_model="d84ap",
        board_id=0x06,
        chip_id=0x8130,
        display_name="iPhone 15 Pro Max",
    ),
    IRecvDevice(
        product_type="iPhone17,3", hardware_model="d47ap", board_id=0x08, chip_id=0x8140, display_name="iPhone 16"
    ),
    IRecvDevice(
        product_type="iPhone17,4", hardware_model="d48ap", board_id=0x0A, chip_id=0x8140, display_name="iPhone 16 Plus"
    ),
    IRecvDevice(
        product_type="iPhone17,5", hardware_model="v59ap", board_id=0x04, chip_id=0x8140, display_name="iPhone 16e"
    ),
    IRecvDevice(
        product_type="iPhone17,1", hardware_model="d93ap", board_id=0x0C, chip_id=0x8140, display_name="iPhone 16 Pro"
    ),
    IRecvDevice(
        product_type="iPhone17,2",
        hardware_model="d94ap",
        board_id=0x0E,
        chip_id=0x8140,
        display_name="iPhone 16 Pro Max",
    ),
    IRecvDevice(
        product_type="iPhone18,4", hardware_model="d23ap", board_id=0x0A, chip_id=0x8150, display_name="iPhone Air"
    ),
    IRecvDevice(
        product_type="iPhone18,1", hardware_model="v53ap", board_id=0x0C, chip_id=0x8150, display_name="iPhone 17 Pro"
    ),
    IRecvDevice(
        product_type="iPhone18,2",
        hardware_model="v54ap",
        board_id=0x0E,
        chip_id=0x8150,
        display_name="iPhone 17 Pro Max",
    ),
    IRecvDevice(
        product_type="iPhone18,3", hardware_model="v57ap", board_id=0x08, chip_id=0x8150, display_name="iPhone 17"
    ),
    IRecvDevice(
        product_type="iPhone18,5", hardware_model="v159ap", board_id=0x16, chip_id=0x8150, display_name="iPhone 17e"
    ),
    # iPod
    IRecvDevice(
        product_type="iPod1,1",
        hardware_model="n45ap",
        board_id=0x02,
        chip_id=0x8900,
        display_name="iPod Touch (1st gen)",
    ),
    IRecvDevice(
        product_type="iPod2,1",
        hardware_model="n72ap",
        board_id=0x00,
        chip_id=0x8720,
        display_name="iPod Touch (2nd gen)",
    ),
    IRecvDevice(
        product_type="iPod3,1",
        hardware_model="n18ap",
        board_id=0x02,
        chip_id=0x8922,
        display_name="iPod Touch (3rd gen)",
    ),
    IRecvDevice(
        product_type="iPod4,1",
        hardware_model="n81ap",
        board_id=0x08,
        chip_id=0x8930,
        display_name="iPod Touch (4th gen)",
    ),
    IRecvDevice(
        product_type="iPod5,1",
        hardware_model="n78ap",
        board_id=0x00,
        chip_id=0x8942,
        display_name="iPod Touch (5th gen)",
    ),
    IRecvDevice(
        product_type="iPod7,1",
        hardware_model="n102ap",
        board_id=0x10,
        chip_id=0x7000,
        display_name="iPod Touch (6th gen)",
    ),
    IRecvDevice(
        product_type="iPod9,1",
        hardware_model="n112ap",
        board_id=0x16,
        chip_id=0x8010,
        display_name="iPod Touch (7th gen)",
    ),
    # iPad
    IRecvDevice(product_type="iPad1,1", hardware_model="k48ap", board_id=0x02, chip_id=0x8930, display_name="iPad"),
    IRecvDevice(
        product_type="iPad2,1", hardware_model="k93ap", board_id=0x04, chip_id=0x8940, display_name="iPad 2 (WiFi)"
    ),
    IRecvDevice(
        product_type="iPad2,2", hardware_model="k94ap", board_id=0x06, chip_id=0x8940, display_name="iPad 2 (GSM)"
    ),
    IRecvDevice(
        product_type="iPad2,3", hardware_model="k95ap", board_id=0x02, chip_id=0x8940, display_name="iPad 2 (CDMA)"
    ),
    IRecvDevice(
        product_type="iPad2,4",
        hardware_model="k93aap",
        board_id=0x06,
        chip_id=0x8942,
        display_name="iPad 2 (WiFi) R2 2012",
    ),
    IRecvDevice(
        product_type="iPad2,5", hardware_model="p105ap", board_id=0x0A, chip_id=0x8942, display_name="iPad mini (WiFi)"
    ),
    IRecvDevice(
        product_type="iPad2,6", hardware_model="p106ap", board_id=0x0C, chip_id=0x8942, display_name="iPad mini (GSM)"
    ),
    IRecvDevice(
        product_type="iPad2,7",
        hardware_model="p107ap",
        board_id=0x0E,
        chip_id=0x8942,
        display_name="iPad mini (Global)",
    ),
    IRecvDevice(
        product_type="iPad3,1",
        hardware_model="j1ap",
        board_id=0x00,
        chip_id=0x8945,
        display_name="iPad (3rd gen, WiFi)",
    ),
    IRecvDevice(
        product_type="iPad3,2",
        hardware_model="j2ap",
        board_id=0x02,
        chip_id=0x8945,
        display_name="iPad (3rd gen, CDMA)",
    ),
    IRecvDevice(
        product_type="iPad3,3",
        hardware_model="j2aap",
        board_id=0x04,
        chip_id=0x8945,
        display_name="iPad (3rd gen, GSM)",
    ),
    IRecvDevice(
        product_type="iPad3,4",
        hardware_model="p101ap",
        board_id=0x00,
        chip_id=0x8955,
        display_name="iPad (4th gen, WiFi)",
    ),
    IRecvDevice(
        product_type="iPad3,5",
        hardware_model="p102ap",
        board_id=0x02,
        chip_id=0x8955,
        display_name="iPad (4th gen, GSM)",
    ),
    IRecvDevice(
        product_type="iPad3,6",
        hardware_model="p103ap",
        board_id=0x04,
        chip_id=0x8955,
        display_name="iPad (4th gen, Global)",
    ),
    IRecvDevice(
        product_type="iPad4,1", hardware_model="j71ap", board_id=0x10, chip_id=0x8960, display_name="iPad Air (WiFi)"
    ),
    IRecvDevice(
        product_type="iPad4,2",
        hardware_model="j72ap",
        board_id=0x12,
        chip_id=0x8960,
        display_name="iPad Air (Cellular)",
    ),
    IRecvDevice(
        product_type="iPad4,3", hardware_model="j73ap", board_id=0x14, chip_id=0x8960, display_name="iPad Air (China)"
    ),
    IRecvDevice(
        product_type="iPad4,4", hardware_model="j85ap", board_id=0x0A, chip_id=0x8960, display_name="iPad mini 2 (WiFi)"
    ),
    IRecvDevice(
        product_type="iPad4,5",
        hardware_model="j86ap",
        board_id=0x0C,
        chip_id=0x8960,
        display_name="iPad mini 2 (Cellular)",
    ),
    IRecvDevice(
        product_type="iPad4,6",
        hardware_model="j87ap",
        board_id=0x0E,
        chip_id=0x8960,
        display_name="iPad mini 2 (China)",
    ),
    IRecvDevice(
        product_type="iPad4,7",
        hardware_model="j85map",
        board_id=0x32,
        chip_id=0x8960,
        display_name="iPad mini 3 (WiFi)",
    ),
    IRecvDevice(
        product_type="iPad4,8",
        hardware_model="j86map",
        board_id=0x34,
        chip_id=0x8960,
        display_name="iPad mini 3 (Cellular)",
    ),
    IRecvDevice(
        product_type="iPad4,9",
        hardware_model="j87map",
        board_id=0x36,
        chip_id=0x8960,
        display_name="iPad mini 3 (China)",
    ),
    IRecvDevice(
        product_type="iPad5,1", hardware_model="j96ap", board_id=0x08, chip_id=0x7000, display_name="iPad mini 4 (WiFi)"
    ),
    IRecvDevice(
        product_type="iPad5,2",
        hardware_model="j97ap",
        board_id=0x0A,
        chip_id=0x7000,
        display_name="iPad mini 4 (Cellular)",
    ),
    IRecvDevice(
        product_type="iPad5,3", hardware_model="j81ap", board_id=0x06, chip_id=0x7001, display_name="iPad Air 2 (WiFi)"
    ),
    IRecvDevice(
        product_type="iPad5,4",
        hardware_model="j82ap",
        board_id=0x02,
        chip_id=0x7001,
        display_name="iPad Air 2 (Cellular)",
    ),
    IRecvDevice(
        product_type="iPad6,3",
        hardware_model="j127ap",
        board_id=0x08,
        chip_id=0x8001,
        display_name="iPad Pro 9.7-inch (WiFi)",
    ),
    IRecvDevice(
        product_type="iPad6,4",
        hardware_model="j128ap",
        board_id=0x0A,
        chip_id=0x8001,
        display_name="iPad Pro 9.7-inch (Cellular)",
    ),
    IRecvDevice(
        product_type="iPad6,7",
        hardware_model="j98aap",
        board_id=0x10,
        chip_id=0x8001,
        display_name="iPad Pro 12.9-inch (1st gen, WiFi)",
    ),
    IRecvDevice(
        product_type="iPad6,8",
        hardware_model="j99aap",
        board_id=0x12,
        chip_id=0x8001,
        display_name="iPad Pro 12.9-inch (1st gen, Cellular)",
    ),
    IRecvDevice(
        product_type="iPad6,11",
        hardware_model="j71sap",
        board_id=0x10,
        chip_id=0x8000,
        display_name="iPad (5th gen, WiFi)",
    ),
    IRecvDevice(
        product_type="iPad6,11",
        hardware_model="j71tap",
        board_id=0x10,
        chip_id=0x8003,
        display_name="iPad (5th gen, WiFi)",
    ),
    IRecvDevice(
        product_type="iPad6,12",
        hardware_model="j72sap",
        board_id=0x12,
        chip_id=0x8000,
        display_name="iPad (5th gen, Cellular)",
    ),
    IRecvDevice(
        product_type="iPad6,12",
        hardware_model="j72tap",
        board_id=0x12,
        chip_id=0x8003,
        display_name="iPad (5th gen, Cellular)",
    ),
    IRecvDevice(
        product_type="iPad7,1",
        hardware_model="j120ap",
        board_id=0x0C,
        chip_id=0x8011,
        display_name="iPad Pro 12.9-inch (2nd gen, WiFi)",
    ),
    IRecvDevice(
        product_type="iPad7,2",
        hardware_model="j121ap",
        board_id=0x0E,
        chip_id=0x8011,
        display_name="iPad Pro 12.9-inch (2nd gen, Cellular)",
    ),
    IRecvDevice(
        product_type="iPad7,3",
        hardware_model="j207ap",
        board_id=0x04,
        chip_id=0x8011,
        display_name="iPad Pro 10.5-inch (WiFi)",
    ),
    IRecvDevice(
        product_type="iPad7,4",
        hardware_model="j208ap",
        board_id=0x06,
        chip_id=0x8011,
        display_name="iPad Pro 10.5-inch (Cellular)",
    ),
    IRecvDevice(
        product_type="iPad7,5",
        hardware_model="j71bap",
        board_id=0x18,
        chip_id=0x8010,
        display_name="iPad (6th gen, WiFi)",
    ),
    IRecvDevice(
        product_type="iPad7,6",
        hardware_model="j72bap",
        board_id=0x1A,
        chip_id=0x8010,
        display_name="iPad (6th gen, Cellular)",
    ),
    IRecvDevice(
        product_type="iPad7,11",
        hardware_model="j171ap",
        board_id=0x1C,
        chip_id=0x8010,
        display_name="iPad (7th gen, WiFi)",
    ),
    IRecvDevice(
        product_type="iPad7,12",
        hardware_model="j172ap",
        board_id=0x1E,
        chip_id=0x8010,
        display_name="iPad (7th gen, Cellular)",
    ),
    IRecvDevice(
        product_type="iPad8,1",
        hardware_model="j317ap",
        board_id=0x0C,
        chip_id=0x8027,
        display_name="iPad Pro 11-inch (1st gen, WiFi)",
    ),
    IRecvDevice(
        product_type="iPad8,2",
        hardware_model="j317xap",
        board_id=0x1C,
        chip_id=0x8027,
        display_name="iPad Pro 11-inch (1st gen, WiFi, 1TB)",
    ),
    IRecvDevice(
        product_type="iPad8,3",
        hardware_model="j318ap",
        board_id=0x0E,
        chip_id=0x8027,
        display_name="iPad Pro 11-inch (1st gen, Cellular)",
    ),
    IRecvDevice(
        product_type="iPad8,4",
        hardware_model="j318xap",
        board_id=0x1E,
        chip_id=0x8027,
        display_name="iPad Pro 11-inch (1st gen, Cellular, 1TB)",
    ),
    IRecvDevice(
        product_type="iPad8,5",
        hardware_model="j320ap",
        board_id=0x08,
        chip_id=0x8027,
        display_name="iPad Pro 12.9-inch (3rd gen, WiFi)",
    ),
    IRecvDevice(
        product_type="iPad8,6",
        hardware_model="j320xap",
        board_id=0x18,
        chip_id=0x8027,
        display_name="iPad Pro 12.9-inch (3rd gen, WiFi, 1TB)",
    ),
    IRecvDevice(
        product_type="iPad8,7",
        hardware_model="j321ap",
        board_id=0x0A,
        chip_id=0x8027,
        display_name="iPad Pro 12.9-inch (3rd gen, Cellular)",
    ),
    IRecvDevice(
        product_type="iPad8,8",
        hardware_model="j321xap",
        board_id=0x1A,
        chip_id=0x8027,
        display_name="iPad Pro 12.9-inch (3rd gen, Cellular, 1TB)",
    ),
    IRecvDevice(
        product_type="iPad8,9",
        hardware_model="j417ap",
        board_id=0x3C,
        chip_id=0x8027,
        display_name="iPad Pro 11-inch (2nd gen, WiFi)",
    ),
    IRecvDevice(
        product_type="iPad8,10",
        hardware_model="j418ap",
        board_id=0x3E,
        chip_id=0x8027,
        display_name="iPad Pro 11-inch (2nd gen, Cellular)",
    ),
    IRecvDevice(
        product_type="iPad8,11",
        hardware_model="j420ap",
        board_id=0x38,
        chip_id=0x8027,
        display_name="iPad Pro 12.9-inch (4th gen, WiFi)",
    ),
    IRecvDevice(
        product_type="iPad8,12",
        hardware_model="j421ap",
        board_id=0x3A,
        chip_id=0x8027,
        display_name="iPad Pro 12.9-inch (4th gen, Cellular)",
    ),
    IRecvDevice(
        product_type="iPad11,1",
        hardware_model="j210ap",
        board_id=0x14,
        chip_id=0x8020,
        display_name="iPad mini (5th gen, WiFi)",
    ),
    IRecvDevice(
        product_type="iPad11,2",
        hardware_model="j211ap",
        board_id=0x16,
        chip_id=0x8020,
        display_name="iPad mini (5th gen, Cellular)",
    ),
    IRecvDevice(
        product_type="iPad11,3",
        hardware_model="j217ap",
        board_id=0x1C,
        chip_id=0x8020,
        display_name="iPad Air (3rd gen, WiFi)",
    ),
    IRecvDevice(
        product_type="iPad11,4",
        hardware_model="j218ap",
        board_id=0x1E,
        chip_id=0x8020,
        display_name="iPad Air (3rd gen, Celluar)",
    ),
    IRecvDevice(
        product_type="iPad11,6",
        hardware_model="j171aap",
        board_id=0x24,
        chip_id=0x8020,
        display_name="iPad (8th gen, WiFi)",
    ),
    IRecvDevice(
        product_type="iPad11,7",
        hardware_model="j172aap",
        board_id=0x26,
        chip_id=0x8020,
        display_name="iPad (8th gen, Cellular)",
    ),
    IRecvDevice(
        product_type="iPad12,1",
        hardware_model="j181ap",
        board_id=0x18,
        chip_id=0x8030,
        display_name="iPad (9th gen, WiFi)",
    ),
    IRecvDevice(
        product_type="iPad12,2",
        hardware_model="j182ap",
        board_id=0x1A,
        chip_id=0x8030,
        display_name="iPad (9th gen, Cellular)",
    ),
    IRecvDevice(
        product_type="iPad13,1",
        hardware_model="j307ap",
        board_id=0x04,
        chip_id=0x8101,
        display_name="iPad Air (4th gen, WiFi)",
    ),
    IRecvDevice(
        product_type="iPad13,2",
        hardware_model="j308ap",
        board_id=0x06,
        chip_id=0x8101,
        display_name="iPad Air (4th gen, Celluar)",
    ),
    IRecvDevice(
        product_type="iPad13,4",
        hardware_model="j517ap",
        board_id=0x08,
        chip_id=0x8103,
        display_name="iPad Pro 11-inch (3rd gen, WiFi)",
    ),
    IRecvDevice(
        product_type="iPad13,5",
        hardware_model="j517xap",
        board_id=0x0A,
        chip_id=0x8103,
        display_name="iPad Pro 11-inch (3rd gen, WiFi, 2TB)",
    ),
    IRecvDevice(
        product_type="iPad13,6",
        hardware_model="j518ap",
        board_id=0x0C,
        chip_id=0x8103,
        display_name="iPad Pro 11-inch (3rd gen, Cellular)",
    ),
    IRecvDevice(
        product_type="iPad13,7",
        hardware_model="j518xap",
        board_id=0x0E,
        chip_id=0x8103,
        display_name="iPad Pro 11-inch (3rd gen, Cellular, 2TB)",
    ),
    IRecvDevice(
        product_type="iPad13,8",
        hardware_model="j522ap",
        board_id=0x18,
        chip_id=0x8103,
        display_name="iPad Pro 12.9-inch (5th gen, WiFi)",
    ),
    IRecvDevice(
        product_type="iPad13,9",
        hardware_model="j522xap",
        board_id=0x1A,
        chip_id=0x8103,
        display_name="iPad Pro 12.9-inch (5th gen, WiFi, 2TB)",
    ),
    IRecvDevice(
        product_type="iPad13,10",
        hardware_model="j523ap",
        board_id=0x1C,
        chip_id=0x8103,
        display_name="iPad Pro 12.9-inch (5th gen, Cellular)",
    ),
    IRecvDevice(
        product_type="iPad13,11",
        hardware_model="j523xap",
        board_id=0x1E,
        chip_id=0x8103,
        display_name="iPad Pro 12.9-inch (5th gen, Cellular, 2TB)",
    ),
    IRecvDevice(
        product_type="iPad13,16",
        hardware_model="j407ap",
        board_id=0x10,
        chip_id=0x8103,
        display_name="iPad Air (5th gen, WiFi)",
    ),
    IRecvDevice(
        product_type="iPad13,17",
        hardware_model="j408ap",
        board_id=0x12,
        chip_id=0x8103,
        display_name="iPad Air (5th gen, Celluar)",
    ),
    IRecvDevice(
        product_type="iPad14,1",
        hardware_model="j310ap",
        board_id=0x04,
        chip_id=0x8110,
        display_name="iPad mini (6th gen, WiFi)",
    ),
    IRecvDevice(
        product_type="iPad14,2",
        hardware_model="j311ap",
        board_id=0x06,
        chip_id=0x8110,
        display_name="iPad mini (6th gen, Cellular)",
    ),
    IRecvDevice(
        product_type="iPad14,3",
        hardware_model="j617ap",
        board_id=0x08,
        chip_id=0x8112,
        display_name="iPad Pro 11-inch (4th gen, WiFi)",
    ),
    IRecvDevice(
        product_type="iPad14,4",
        hardware_model="j618ap",
        board_id=0x0A,
        chip_id=0x8112,
        display_name="iPad Pro 11-inch (4th gen, Cellular)",
    ),
    IRecvDevice(
        product_type="iPad14,5",
        hardware_model="j620ap",
        board_id=0x0C,
        chip_id=0x8112,
        display_name="iPad Pro 12.9-inch (6th gen, WiFi)",
    ),
    IRecvDevice(
        product_type="iPad14,6",
        hardware_model="j621ap",
        board_id=0x0E,
        chip_id=0x8112,
        display_name="iPad Pro 12.9-inch (6th gen, Cellular)",
    ),
    IRecvDevice(
        product_type="iPad14,8",
        hardware_model="j507ap",
        board_id=0x10,
        chip_id=0x8112,
        display_name="iPad Air 11-inch (M2, WiFi)",
    ),
    IRecvDevice(
        product_type="iPad14,9",
        hardware_model="j508ap",
        board_id=0x12,
        chip_id=0x8112,
        display_name="iPad Air 11-inch (M2, Celluar)",
    ),
    IRecvDevice(
        product_type="iPad14,10",
        hardware_model="j537ap",
        board_id=0x14,
        chip_id=0x8112,
        display_name="iPad Air 13-inch (M2, WiFi)",
    ),
    IRecvDevice(
        product_type="iPad14,11",
        hardware_model="j538ap",
        board_id=0x16,
        chip_id=0x8112,
        display_name="iPad Air 13-inch (M2, Celluar)",
    ),
    IRecvDevice(
        product_type="iPad15,3",
        hardware_model="j607ap",
        board_id=0x08,
        chip_id=0x8122,
        display_name="iPad Air 11-inch (M3, WiFi)",
    ),
    IRecvDevice(
        product_type="iPad15,4",
        hardware_model="j608ap",
        board_id=0x0A,
        chip_id=0x8122,
        display_name="iPad Air 11-inch (M3, Celluar)",
    ),
    IRecvDevice(
        product_type="iPad15,5",
        hardware_model="j637ap",
        board_id=0x0C,
        chip_id=0x8122,
        display_name="iPad Air 13-inch (M3, WiFi)",
    ),
    IRecvDevice(
        product_type="iPad15,6",
        hardware_model="j638ap",
        board_id=0x0E,
        chip_id=0x8122,
        display_name="iPad Air 13-inch (M3, Celluar)",
    ),
    IRecvDevice(
        product_type="iPad15,7",
        hardware_model="j481ap",
        board_id=0x10,
        chip_id=0x8120,
        display_name="iPad (A16, WiFi)",
    ),
    IRecvDevice(
        product_type="iPad15,8",
        hardware_model="j482ap",
        board_id=0x12,
        chip_id=0x8120,
        display_name="iPad (A16, Celluar)",
    ),
    IRecvDevice(
        product_type="iPad16,1",
        hardware_model="j410ap",
        board_id=0x08,
        chip_id=0x8130,
        display_name="iPad mini (A17 Pro, WiFi)",
    ),
    IRecvDevice(
        product_type="iPad16,2",
        hardware_model="j411ap",
        board_id=0x0A,
        chip_id=0x8130,
        display_name="iPad mini (A17 Pro, Celluar)",
    ),
    IRecvDevice(
        product_type="iPad16,3",
        hardware_model="j717ap",
        board_id=0x08,
        chip_id=0x8132,
        display_name="iPad Pro 11-inch (M4, WiFi)",
    ),
    IRecvDevice(
        product_type="iPad16,4",
        hardware_model="j718ap",
        board_id=0x0A,
        chip_id=0x8132,
        display_name="iPad Pro 11-inch (M4, Cellular)",
    ),
    IRecvDevice(
        product_type="iPad16,5",
        hardware_model="j720ap",
        board_id=0x0C,
        chip_id=0x8132,
        display_name="iPad Pro 13-inch (M4, WiFi)",
    ),
    IRecvDevice(
        product_type="iPad16,6",
        hardware_model="j721ap",
        board_id=0x0E,
        chip_id=0x8132,
        display_name="iPad Pro 13-inch (M4, Cellular)",
    ),
    IRecvDevice(
        product_type="iPad16,8",
        hardware_model="j707ap",
        board_id=0x10,
        chip_id=0x8132,
        display_name="iPad Air 11-inch (M4, WiFi)",
    ),
    IRecvDevice(
        product_type="iPad16,9",
        hardware_model="j708ap",
        board_id=0x12,
        chip_id=0x8132,
        display_name="iPad Air 11-inch (M4, Celluar)",
    ),
    IRecvDevice(
        product_type="iPad16,10",
        hardware_model="j737ap",
        board_id=0x14,
        chip_id=0x8132,
        display_name="iPad Air 13-inch (M4, WiFi)",
    ),
    IRecvDevice(
        product_type="iPad16,11",
        hardware_model="j738ap",
        board_id=0x16,
        chip_id=0x8132,
        display_name="iPad Air 13-inch (M4, Celluar)",
    ),
    # Apple TV
    IRecvDevice(
        product_type="AppleTV2,1", hardware_model="k66ap", board_id=0x10, chip_id=0x8930, display_name="Apple TV 2"
    ),
    IRecvDevice(
        product_type="AppleTV3,1", hardware_model="j33ap", board_id=0x08, chip_id=0x8942, display_name="Apple TV 3"
    ),
    IRecvDevice(
        product_type="AppleTV3,2",
        hardware_model="j33iap",
        board_id=0x00,
        chip_id=0x8947,
        display_name="Apple TV 3 (2013)",
    ),
    IRecvDevice(
        product_type="AppleTV5,3", hardware_model="j42dap", board_id=0x34, chip_id=0x7000, display_name="Apple TV 4"
    ),
    IRecvDevice(
        product_type="AppleTV6,2", hardware_model="j105aap", board_id=0x02, chip_id=0x8011, display_name="Apple TV 4K"
    ),
    IRecvDevice(
        product_type="AppleTV11,1",
        hardware_model="j305ap",
        board_id=0x08,
        chip_id=0x8020,
        display_name="Apple TV 4K (2nd gen)",
    ),
    # HomePod
    IRecvDevice(
        product_type="AudioAccessory1,1",
        hardware_model="b238aap",
        board_id=0x38,
        chip_id=0x7000,
        display_name="HomePod",
    ),
    IRecvDevice(
        product_type="AudioAccessory1,2", hardware_model="b238ap", board_id=0x1A, chip_id=0x7000, display_name="HomePod"
    ),
    IRecvDevice(
        product_type="AudioAccessory5,1",
        hardware_model="b520ap",
        board_id=0x22,
        chip_id=0x8006,
        display_name="HomePod mini",
    ),
    # Apple Watch
    IRecvDevice(
        product_type="Watch1,1",
        hardware_model="n27aap",
        board_id=0x02,
        chip_id=0x7002,
        display_name="Apple Watch 38mm (1st gen)",
    ),
    IRecvDevice(
        product_type="Watch1,2",
        hardware_model="n28aap",
        board_id=0x04,
        chip_id=0x7002,
        display_name="Apple Watch 42mm (1st gen)",
    ),
    IRecvDevice(
        product_type="Watch2,6",
        hardware_model="n27dap",
        board_id=0x02,
        chip_id=0x8002,
        display_name="Apple Watch Series 1 (38mm)",
    ),
    IRecvDevice(
        product_type="Watch2,7",
        hardware_model="n28dap",
        board_id=0x04,
        chip_id=0x8002,
        display_name="Apple Watch Series 1 (42mm)",
    ),
    IRecvDevice(
        product_type="Watch2,3",
        hardware_model="n74ap",
        board_id=0x0C,
        chip_id=0x8002,
        display_name="Apple Watch Series 2 (38mm)",
    ),
    IRecvDevice(
        product_type="Watch2,4",
        hardware_model="n75ap",
        board_id=0x0E,
        chip_id=0x8002,
        display_name="Apple Watch Series 2 (42mm)",
    ),
    IRecvDevice(
        product_type="Watch3,1",
        hardware_model="n111sap",
        board_id=0x1C,
        chip_id=0x8004,
        display_name="Apple Watch Series 3 (38mm Cellular)",
    ),
    IRecvDevice(
        product_type="Watch3,2",
        hardware_model="n111bap",
        board_id=0x1E,
        chip_id=0x8004,
        display_name="Apple Watch Series 3 (42mm Cellular)",
    ),
    IRecvDevice(
        product_type="Watch3,3",
        hardware_model="n121sap",
        board_id=0x18,
        chip_id=0x8004,
        display_name="Apple Watch Series 3 (38mm)",
    ),
    IRecvDevice(
        product_type="Watch3,4",
        hardware_model="n121bap",
        board_id=0x1A,
        chip_id=0x8004,
        display_name="Apple Watch Series 3 (42mm)",
    ),
    IRecvDevice(
        product_type="Watch4,1",
        hardware_model="n131sap",
        board_id=0x08,
        chip_id=0x8006,
        display_name="Apple Watch Series 4 (40mm)",
    ),
    IRecvDevice(
        product_type="Watch4,2",
        hardware_model="n131bap",
        board_id=0x0A,
        chip_id=0x8006,
        display_name="Apple Watch Series 4 (44mm)",
    ),
    IRecvDevice(
        product_type="Watch4,3",
        hardware_model="n141sap",
        board_id=0x0C,
        chip_id=0x8006,
        display_name="Apple Watch Series 4 (40mm Cellular)",
    ),
    IRecvDevice(
        product_type="Watch4,4",
        hardware_model="n141bap",
        board_id=0x0E,
        chip_id=0x8006,
        display_name="Apple Watch Series 4 (44mm Cellular)",
    ),
    IRecvDevice(
        product_type="Watch5,1",
        hardware_model="n144sap",
        board_id=0x10,
        chip_id=0x8006,
        display_name="Apple Watch Series 5 (40mm)",
    ),
    IRecvDevice(
        product_type="Watch5,2",
        hardware_model="n144bap",
        board_id=0x12,
        chip_id=0x8006,
        display_name="Apple Watch Series 5 (44mm)",
    ),
    IRecvDevice(
        product_type="Watch5,3",
        hardware_model="n146sap",
        board_id=0x14,
        chip_id=0x8006,
        display_name="Apple Watch Series 5 (40mm Cellular)",
    ),
    IRecvDevice(
        product_type="Watch5,4",
        hardware_model="n146bap",
        board_id=0x16,
        chip_id=0x8006,
        display_name="Apple Watch Series 5 (44mm Cellular)",
    ),
    IRecvDevice(
        product_type="Watch5,9",
        hardware_model="n140sap",
        board_id=0x28,
        chip_id=0x8006,
        display_name="Apple Watch SE (40mm)",
    ),
    IRecvDevice(
        product_type="Watch5,10",
        hardware_model="n140bap",
        board_id=0x2A,
        chip_id=0x8006,
        display_name="Apple Watch SE (44mm)",
    ),
    IRecvDevice(
        product_type="Watch5,11",
        hardware_model="n142sap",
        board_id=0x2C,
        chip_id=0x8006,
        display_name="Apple Watch SE (40mm Cellular)",
    ),
    IRecvDevice(
        product_type="Watch5,12",
        hardware_model="n142bap",
        board_id=0x2E,
        chip_id=0x8006,
        display_name="Apple Watch SE (44mm Cellular)",
    ),
    IRecvDevice(
        product_type="Watch6,1",
        hardware_model="n157sap",
        board_id=0x08,
        chip_id=0x8301,
        display_name="Apple Watch Series 6 (40mm)",
    ),
    IRecvDevice(
        product_type="Watch6,2",
        hardware_model="n157bap",
        board_id=0x0A,
        chip_id=0x8301,
        display_name="Apple Watch Series 6 (44mm)",
    ),
    IRecvDevice(
        product_type="Watch6,3",
        hardware_model="n158sap",
        board_id=0x0C,
        chip_id=0x8301,
        display_name="Apple Watch Series 6 (40mm Cellular)",
    ),
    IRecvDevice(
        product_type="Watch6,4",
        hardware_model="n158bap",
        board_id=0x0E,
        chip_id=0x8301,
        display_name="Apple Watch Series 6 (44mm Cellular)",
    ),
    IRecvDevice(
        product_type="Watch6,6",
        hardware_model="n187sap",
        board_id=0x10,
        chip_id=0x8301,
        display_name="Apple Watch Series 7 (41mm)",
    ),
    IRecvDevice(
        product_type="Watch6,7",
        hardware_model="n187bap",
        board_id=0x12,
        chip_id=0x8301,
        display_name="Apple Watch Series 7 (45mm)",
    ),
    IRecvDevice(
        product_type="Watch6,8",
        hardware_model="n188sap",
        board_id=0x14,
        chip_id=0x8301,
        display_name="Apple Watch Series 7 (41mm Cellular)",
    ),
    IRecvDevice(
        product_type="Watch6,9",
        hardware_model="n188bap",
        board_id=0x16,
        chip_id=0x8301,
        display_name="Apple Watch Series 7 (45mm Cellular)",
    ),
    IRecvDevice(
        product_type="Watch6,10",
        hardware_model="n143sap",
        board_id=0x18,
        chip_id=0x8301,
        display_name="Apple Watch SE (2nd generation) (40mm)",
    ),
    IRecvDevice(
        product_type="Watch6,11",
        hardware_model="n143bap",
        board_id=0x2A,
        chip_id=0x8301,
        display_name="Apple Watch SE (2nd generation) (44mm)",
    ),
    IRecvDevice(
        product_type="Watch6,12",
        hardware_model="n149sap",
        board_id=0x2C,
        chip_id=0x8301,
        display_name="Apple Watch SE (2nd generation) (40mm Cellular)",
    ),
    IRecvDevice(
        product_type="Watch6,13",
        hardware_model="n149bap",
        board_id=0x2E,
        chip_id=0x8301,
        display_name="Apple Watch SE (2nd generation) (44m Cellular)",
    ),
    IRecvDevice(
        product_type="Watch6,14",
        hardware_model="n197sap",
        board_id=0x30,
        chip_id=0x8301,
        display_name="Apple Watch Series 8 (41mm)",
    ),
    IRecvDevice(
        product_type="Watch6,15",
        hardware_model="n197bap",
        board_id=0x32,
        chip_id=0x8301,
        display_name="Apple Watch Series 8 (45mm)",
    ),
    IRecvDevice(
        product_type="Watch6,16",
        hardware_model="n198sap",
        board_id=0x34,
        chip_id=0x8301,
        display_name="Apple Watch Series 8 (41mm Cellular)",
    ),
    IRecvDevice(
        product_type="Watch6,17",
        hardware_model="n198bap",
        board_id=0x36,
        chip_id=0x8301,
        display_name="Apple Watch Series 8 (45mm Cellular)",
    ),
    IRecvDevice(
        product_type="Watch6,17",
        hardware_model="n199ap",
        board_id=0x26,
        chip_id=0x8301,
        display_name="Apple Watch Ultra",
    ),
    IRecvDevice(
        product_type="Watch7,1",
        hardware_model="n207sap",
        board_id=0x08,
        chip_id=0x8310,
        display_name="Apple Watch Series 9 (41mm)",
    ),
    IRecvDevice(
        product_type="Watch7,2",
        hardware_model="n207bap",
        board_id=0x0A,
        chip_id=0x8310,
        display_name="Apple Watch Series 9 (45mm)",
    ),
    IRecvDevice(
        product_type="Watch7,3",
        hardware_model="n208sap",
        board_id=0x0C,
        chip_id=0x8310,
        display_name="Apple Watch Series 9 (41mm Cellular)",
    ),
    IRecvDevice(
        product_type="Watch7,4",
        hardware_model="n208bap",
        board_id=0x0E,
        chip_id=0x8310,
        display_name="Apple Watch Series 9 (45mm Cellular)",
    ),
    IRecvDevice(
        product_type="Watch7,5",
        hardware_model="n210ap",
        board_id=0x02,
        chip_id=0x8310,
        display_name="Apple Watch Ultra 2",
    ),
    IRecvDevice(
        product_type="Watch7,8",
        hardware_model="n217sap",
        board_id=0x10,
        chip_id=0x8310,
        display_name="Apple Watch Series 10 (42mm)",
    ),
    IRecvDevice(
        product_type="Watch7,9",
        hardware_model="n217bap",
        board_id=0x12,
        chip_id=0x8310,
        display_name="Apple Watch Series 10 (46mm)",
    ),
    IRecvDevice(
        product_type="Watch7,10",
        hardware_model="n218sap",
        board_id=0x14,
        chip_id=0x8310,
        display_name="Apple Watch Series 10 (42mm Cellular)",
    ),
    IRecvDevice(
        product_type="Watch7,11",
        hardware_model="n218bap",
        board_id=0x16,
        chip_id=0x8310,
        display_name="Apple Watch Series 10 (46mm Cellular)",
    ),
    IRecvDevice(
        product_type="Watch7,12",
        hardware_model="n230ap",
        board_id=0x22,
        chip_id=0x8310,
        display_name="Apple Watch Ultra 3",
    ),
    IRecvDevice(
        product_type="Watch7,13",
        hardware_model="n243sap",
        board_id=0x28,
        chip_id=0x8310,
        display_name="Apple Watch SE 3 (40mm)",
    ),
    IRecvDevice(
        product_type="Watch7,14",
        hardware_model="n243bap",
        board_id=0x2A,
        chip_id=0x8310,
        display_name="Apple Watch SE 3 (44mm)",
    ),
    IRecvDevice(
        product_type="Watch7,15",
        hardware_model="n244sap",
        board_id=0x2C,
        chip_id=0x8310,
        display_name="Apple Watch SE 3 (40mm Cellular)",
    ),
    IRecvDevice(
        product_type="Watch7,16",
        hardware_model="n244bap",
        board_id=0x2E,
        chip_id=0x8310,
        display_name="Apple Watch SE 3 (44mm Cellular)",
    ),
    IRecvDevice(
        product_type="Watch7,17",
        hardware_model="n227sap",
        board_id=0x18,
        chip_id=0x8310,
        display_name="Apple Watch Series 11 (42mm)",
    ),
    IRecvDevice(
        product_type="Watch7,18",
        hardware_model="n227bap",
        board_id=0x1A,
        chip_id=0x8310,
        display_name="Apple Watch Series 11 (46mm)",
    ),
    IRecvDevice(
        product_type="Watch7,19",
        hardware_model="n228sap",
        board_id=0x1C,
        chip_id=0x8310,
        display_name="Apple Watch Series 11 (42mm Cellular)",
    ),
    IRecvDevice(
        product_type="Watch7,20",
        hardware_model="n228bap",
        board_id=0x1E,
        chip_id=0x8310,
        display_name="Apple Watch Series 11 (46mm Cellular)",
    ),
    # Apple Silicon Macs
    IRecvDevice(
        product_type="ADP3,2",
        hardware_model="j273aap",
        board_id=0x42,
        chip_id=0x8027,
        display_name="Developer Transition Kit (2020)",
    ),
    IRecvDevice(
        product_type="Macmini9,1",
        hardware_model="j274ap",
        board_id=0x22,
        chip_id=0x8103,
        display_name="Mac mini (M1, 2020)",
    ),
    IRecvDevice(
        product_type="MacBookPro17,1",
        hardware_model="j293ap",
        board_id=0x24,
        chip_id=0x8103,
        display_name="MacBook Pro (M1, 13-inch, 2020)",
    ),
    IRecvDevice(
        product_type="MacBookPro18,1",
        hardware_model="j316sap",
        board_id=0x0A,
        chip_id=0x6000,
        display_name="MacBook Pro (M1 Pro, 16-inch, 2021)",
    ),
    IRecvDevice(
        product_type="MacBookPro18,2",
        hardware_model="j316cap",
        board_id=0x0A,
        chip_id=0x6001,
        display_name="MacBook Pro (M1 Max, 16-inch, 2021)",
    ),
    IRecvDevice(
        product_type="MacBookPro18,3",
        hardware_model="j314sap",
        board_id=0x08,
        chip_id=0x6000,
        display_name="MacBook Pro (M1 Pro, 14-inch, 2021)",
    ),
    IRecvDevice(
        product_type="MacBookPro18,4",
        hardware_model="j314cap",
        board_id=0x08,
        chip_id=0x6001,
        display_name="MacBook Pro (M1 Max, 14-inch, 2021)",
    ),
    IRecvDevice(
        product_type="MacBookAir10,1",
        hardware_model="j313ap",
        board_id=0x26,
        chip_id=0x8103,
        display_name="MacBook Air (M1, 2020)",
    ),
    IRecvDevice(
        product_type="iMac21,1",
        hardware_model="j456ap",
        board_id=0x28,
        chip_id=0x8103,
        display_name="iMac 24-inch (M1, Two Ports, 2021)",
    ),
    IRecvDevice(
        product_type="iMac21,2",
        hardware_model="j457ap",
        board_id=0x2A,
        chip_id=0x8103,
        display_name="iMac 24-inch (M1, Four Ports, 2021)",
    ),
    IRecvDevice(
        product_type="Mac13,1",
        hardware_model="j375cap",
        board_id=0x04,
        chip_id=0x6001,
        display_name="Mac Studio (M1 Max, 2022)",
    ),
    IRecvDevice(
        product_type="Mac13,2",
        hardware_model="j375dap",
        board_id=0x0C,
        chip_id=0x6002,
        display_name="Mac Studio (M1 Ultra, 2022)",
    ),
    IRecvDevice(
        product_type="Mac14,2",
        hardware_model="j413ap",
        board_id=0x28,
        chip_id=0x8112,
        display_name="MacBook Air (M2, 2022)",
    ),
    IRecvDevice(
        product_type="Mac14,7",
        hardware_model="j493ap",
        board_id=0x2A,
        chip_id=0x8112,
        display_name="MacBook Pro (M2, 13-inch, 2022)",
    ),
    # Apple T2 Coprocessor
    IRecvDevice(
        product_type="iBridge2,1",
        hardware_model="j137ap",
        board_id=0x0A,
        chip_id=0x8012,
        display_name="Apple T2 iMacPro1,1 (j137)",
    ),
    IRecvDevice(
        product_type="iBridge2,3",
        hardware_model="j680ap",
        board_id=0x0B,
        chip_id=0x8012,
        display_name="Apple T2 MacBookPro15,1 (j680)",
    ),
    IRecvDevice(
        product_type="iBridge2,4",
        hardware_model="j132ap",
        board_id=0x0C,
        chip_id=0x8012,
        display_name="Apple T2 MacBookPro15,2 (j132)",
    ),
    IRecvDevice(
        product_type="iBridge2,5",
        hardware_model="j174ap",
        board_id=0x0E,
        chip_id=0x8012,
        display_name="Apple T2 Macmini8,1 (j174)",
    ),
    IRecvDevice(
        product_type="iBridge2,6",
        hardware_model="j160ap",
        board_id=0x0F,
        chip_id=0x8012,
        display_name="Apple T2 MacPro7,1 (j160)",
    ),
    IRecvDevice(
        product_type="iBridge2,7",
        hardware_model="j780ap",
        board_id=0x07,
        chip_id=0x8012,
        display_name="Apple T2 MacBookPro15,3 (j780)",
    ),
    IRecvDevice(
        product_type="iBridge2,8",
        hardware_model="j140kap",
        board_id=0x17,
        chip_id=0x8012,
        display_name="Apple T2 MacBookAir8,1 (j140k)",
    ),
    IRecvDevice(
        product_type="iBridge2,10",
        hardware_model="j213ap",
        board_id=0x18,
        chip_id=0x8012,
        display_name="Apple T2 MacBookPro15,4 (j213)",
    ),
    IRecvDevice(
        product_type="iBridge2,12",
        hardware_model="j140aap",
        board_id=0x37,
        chip_id=0x8012,
        display_name="Apple T2 MacBookAir8,2 (j140a)",
    ),
    IRecvDevice(
        product_type="iBridge2,14",
        hardware_model="j152fap",
        board_id=0x3A,
        chip_id=0x8012,
        display_name="Apple T2 MacBookPro16,1 (j152f)",
    ),
    IRecvDevice(
        product_type="iBridge2,15",
        hardware_model="j230kap",
        board_id=0x3F,
        chip_id=0x8012,
        display_name="Apple T2 MacBookAir9,1 (j230k)",
    ),
    IRecvDevice(
        product_type="iBridge2,16",
        hardware_model="j214kap",
        board_id=0x3E,
        chip_id=0x8012,
        display_name="Apple T2 MacBookPro16,2 (j214k)",
    ),
    IRecvDevice(
        product_type="iBridge2,19",
        hardware_model="j185ap",
        board_id=0x22,
        chip_id=0x8012,
        display_name="Apple T2 iMac20,1 (j185)",
    ),
    IRecvDevice(
        product_type="iBridge2,20",
        hardware_model="j185fap",
        board_id=0x23,
        chip_id=0x8012,
        display_name="Apple T2 iMac20,2 (j185f)",
    ),
    IRecvDevice(
        product_type="iBridge2,21",
        hardware_model="j223ap",
        board_id=0x3B,
        chip_id=0x8012,
        display_name="Apple T2 MacBookPro16,3 (j223)",
    ),
    IRecvDevice(
        product_type="iBridge2,22",
        hardware_model="j215ap",
        board_id=0x38,
        chip_id=0x8012,
        display_name="Apple T2 MacBookPro16,4 (j215)",
    ),
    # Apple Displays
    IRecvDevice(
        product_type="AppleDisplay2,1",
        hardware_model="j327ap",
        board_id=0x22,
        chip_id=0x8030,
        display_name="Studio Display",
    ),
    # Apple Vision Pro
    IRecvDevice(
        product_type="RealityDevice14,1",
        hardware_model="n301ap",
        board_id=0x42,
        chip_id=0x8112,
        display_name="Apple Vision Pro",
    ),
    IRecvDevice(
        product_type="RealityDevice17,1",
        hardware_model="n301aap",
        board_id=0x42,
        chip_id=0x8142,
        display_name="Apple Vision Pro (M5)",
    ),
    # Virtual iPhone Research Environment
    IRecvDevice(
        product_type="iPhone99,11",
        hardware_model="vresearch101ap",
        board_id=0x90,
        chip_id=0xFE01,
        display_name="Virtual iPhone (99,11)",
    ),
)



"""Tests for DTXFragmenter: assembly, fragmentation, zero-copy, and edge cases."""

from __future__ import annotations

import itertools

import pytest

from pymobiledevice3.dtx.exceptions import DTXProtocolError
from pymobiledevice3.dtx.fragment import DTXFragment
from pymobiledevice3.dtx.fragmenter import MAX_FRAGMENT_SIZE, MAX_MESSAGE_SIZE, DTXFragmenter

# ---------------------------------------------------------------------------
# Helpers
# ---------------------------------------------------------------------------


def _first_frag(total_size: int, count: int = 2, identifier: int = 1) -> DTXFragment:
    """Build a first (header-only) fragment declaring *total_size* bytes."""
    return DTXFragment(
        index=0,
        count=count,
        data_size=total_size,
        identifier=identifier,
    )


def _body_frag(index: int, payload: bytes | bytearray, count: int = 2, identifier: int = 1) -> DTXFragment:
    """Build a non-first body fragment carrying *payload*."""
    return DTXFragment(
        index=index,
        count=count,
        data_size=len(payload),
        identifier=identifier,
        payload=memoryview(bytearray(payload)),
    )


async def _collect(gen) -> list[DTXFragment]:
    """Drain an async generator of DTXFragments into a list."""
    return [f async for f in gen]


# ---------------------------------------------------------------------------
# DTXFragmenter - construction
# ---------------------------------------------------------------------------


class TestDTXFragmenterConstruction:
    def test_basic(self):
        first = _first_frag(total_size=100, count=3)
        fragmenter = DTXFragmenter(first, current_buffered=0, max_buffered_size=1024)
        assert fragmenter.identifier == 1
        assert fragmenter.declared_size == 100

    def test_zero_data_size_raises(self):
        first = _first_frag(total_size=0, count=2)
        with pytest.raises(DTXProtocolError, match="data_size=0"):
            DTXFragmenter(first, current_buffered=0, max_buffered_size=1024)

    def test_exceeds_max_message_size_raises(self):
        first = _first_frag(total_size=MAX_MESSAGE_SIZE + 1, count=2)
        with pytest.raises(DTXProtocolError, match="MAX_MESSAGE_SIZE"):
            DTXFragmenter(first, current_buffered=0, max_buffered_size=MAX_MESSAGE_SIZE * 2)

    def test_exceeds_max_buffered_size_raises(self):
        first = _first_frag(total_size=500, count=2)
        with pytest.raises(DTXProtocolError, match="MAX_BUFFERED_SIZE"):
            DTXFragmenter(first, current_buffered=600, max_buffered_size=1000)

    def test_exactly_at_max_buffered_size_is_ok(self):
        first = _first_frag(total_size=400, count=2)
        # current_buffered + total == max_buffered_size → accepted
        fragmenter = DTXFragmenter(first, current_buffered=600, max_buffered_size=1000)
        assert fragmenter.declared_size == 400

    def test_exactly_at_max_message_size_is_ok(self):
        first = _first_frag(total_size=MAX_MESSAGE_SIZE, count=2)
        fragmenter = DTXFragmenter(first, current_buffered=0, max_buffered_size=MAX_MESSAGE_SIZE)
        assert fragmenter.declared_size == MAX_MESSAGE_SIZE


# ---------------------------------------------------------------------------
# DTXFragmenter - add() / assemble() - in-order (zero-copy path)
# ---------------------------------------------------------------------------


class TestDTXFragmenterInOrder:
    def test_two_fragments_in_order(self):
        body = b"hello world"
        first = _first_frag(total_size=len(body), count=2)
        f = DTXFragmenter(first, current_buffered=0, max_buffered_size=1024)
        internal_buffer = f._buffer  # keep reference to verify zero-copy

        done = f.add(_body_frag(index=1, payload=body))
        assert done is True

        result, meta = f.assemble()
        assert bytes(result) == body
        assert meta is first

        # Zero-copy: in-order fragments must return the pre-allocated buffer,
        # not a new allocation.
        assert result is internal_buffer

    def test_three_fragments_in_order(self):
        part_a = b"AAAA"
        part_b = b"BBBB"
        first = _first_frag(total_size=len(part_a) + len(part_b), count=3)
        f = DTXFragmenter(first, current_buffered=0, max_buffered_size=1024)
        internal_buffer = f._buffer

        assert f.add(_body_frag(index=1, payload=part_a, count=3)) is False
        assert f.add(_body_frag(index=2, payload=part_b, count=3)) is True

        result, _ = f.assemble()
        assert bytes(result) == part_a + part_b
        assert result is internal_buffer

    def test_add_returns_false_until_last(self):
        parts = [bytes([i] * 10) for i in range(1, 5)]
        total = sum(len(p) for p in parts)
        first = _first_frag(total_size=total, count=len(parts) + 1)
        f = DTXFragmenter(first, current_buffered=0, max_buffered_size=total + 100)

        for i, part in enumerate(parts, start=1):
            is_last = i == len(parts)
            assert f.add(_body_frag(index=i, payload=part, count=len(parts) + 1)) is is_last

        result, _ = f.assemble()
        assert bytes(result) == b"".join(parts)


# ---------------------------------------------------------------------------
# DTXFragmenter - assemble() - out-of-order (copy path)
# ---------------------------------------------------------------------------


class TestDTXFragmenterOutOfOrder:
    def test_two_fragments_reversed(self):
        part_a = b"FIRST_PART_"
        part_b = b"SECOND_PART"
        first = _first_frag(total_size=len(part_a) + len(part_b), count=3)
        f = DTXFragmenter(first, current_buffered=0, max_buffered_size=1024)
        original_buffer = f._buffer

        # Arrive in reverse order: index 2 before index 1
        assert f.add(_body_frag(index=2, payload=part_b, count=3)) is False
        assert f.add(_body_frag(index=1, payload=part_a, count=3)) is True

        result, _ = f.assemble()
        assert bytes(result) == part_a + part_b

        # Out-of-order must produce a NEW buffer (the original was written
        # in arrival order, so its bytes are scrambled).
        assert result is not original_buffer

    def test_three_fragments_permutation(self):
        """Every non-identity permutation of 3 body fragments."""
        parts = [bytes([i] * 8) for i in range(1, 4)]
        expected = b"".join(parts)
        total = len(expected)

        for perm in itertools.permutations([1, 2, 3]):
            if perm == (1, 2, 3):
                continue  # in-order covered elsewhere
            first = _first_frag(total_size=total, count=4)
            f = DTXFragmenter(first, current_buffered=0, max_buffered_size=total + 100)
            for idx in perm:
                f.add(_body_frag(index=idx, payload=parts[idx - 1], count=4))
            result, _ = f.assemble()
            assert bytes(result) == expected, f"Failed for permutation {perm}"


# ---------------------------------------------------------------------------
# DTXFragmenter - add() error cases
# ---------------------------------------------------------------------------


class TestDTXFragmenterErrors:
    def test_duplicate_fragment_index_raises(self):
        first = _first_frag(total_size=20, count=3)
        f = DTXFragmenter(first, current_buffered=0, max_buffered_size=1024)
        f.add(_body_frag(index=1, payload=b"A" * 10, count=3))
        with pytest.raises(DTXProtocolError, match=r"[Dd]uplicate"):
            f.add(_body_frag(index=1, payload=b"B" * 5, count=3))

    def test_overflow_raises(self):
        """A fragment whose payload would push the write pointer past declared_size."""
        first = _first_frag(total_size=10, count=2)
        f = DTXFragmenter(first, current_buffered=0, max_buffered_size=1024)
        with pytest.raises(DTXProtocolError, match="exceed"):
            f.add(_body_frag(index=1, payload=b"X" * 11, count=2))

    def test_partial_overflow_raises(self):
        """First fragment OK, second pushes past the boundary."""
        first = _first_frag(total_size=10, count=3)
        f = DTXFragmenter(first, current_buffered=0, max_buffered_size=1024)
        f.add(_body_frag(index=1, payload=b"A" * 6, count=3))
        with pytest.raises(DTXProtocolError, match="exceed"):
            f.add(_body_frag(index=2, payload=b"B" * 5, count=3))  # 6+5=11 > 10

    def test_assemble_before_all_fragments_raises(self):
        first = _first_frag(total_size=20, count=3)
        f = DTXFragmenter(first, current_buffered=0, max_buffered_size=1024)
        f.add(_body_frag(index=1, payload=b"A" * 10, count=3))
        with pytest.raises(AssertionError):
            f.assemble()

    def test_none_payload_raises(self):
        first = _first_frag(total_size=10, count=2)
        f = DTXFragmenter(first, current_buffered=0, max_buffered_size=1024)
        frag = DTXFragment(index=1, count=2, data_size=10, identifier=1, payload=None)
        with pytest.raises(DTXProtocolError):
            f.add(frag)


# ---------------------------------------------------------------------------
# DTXFragmenter.fragment() - single-fragment path
# ---------------------------------------------------------------------------


class TestFragmentSingleFragment:
    async def test_empty_payload(self):
        frags = await _collect(DTXFragmenter.fragment(memoryview(b"")))
        assert len(frags) == 1
        assert frags[0].index == 0
        assert frags[0].count == 1
        assert frags[0].data_size == 0
        assert len(frags[0].payload) == 0

    async def test_single_chunk_zero_copy(self):
        """Single memoryview ≤ MAX_FRAGMENT_SIZE → payload IS the same buffer."""
        src = bytearray(b"the quick brown fox")
        mv = memoryview(src)
        frags = await _collect(DTXFragmenter.fragment(mv))
        assert len(frags) == 1
        frag = frags[0]
        assert frag.index == 0
        assert frag.count == 1
        assert bytes(frag.payload) == bytes(src)
        # The payload must share the same backing bytearray (zero-copy).
        assert frag.payload.obj is src

    async def test_multiple_chunks_fit_in_one_fragment(self):
        """Multiple memoryviews whose total is ≤ MAX_FRAGMENT_SIZE → concatenated once."""
        a = bytearray(b"hello ")
        b_ = bytearray(b"world")
        frags = await _collect(DTXFragmenter.fragment(memoryview(a), memoryview(b_)))
        assert len(frags) == 1
        assert bytes(frags[0].payload) == b"hello world"

    async def test_exactly_max_fragment_size(self):
        src = bytearray(MAX_FRAGMENT_SIZE)
        frags = await _collect(DTXFragmenter.fragment(memoryview(src)))
        assert len(frags) == 1
        assert frags[0].data_size == MAX_FRAGMENT_SIZE
        assert frags[0].payload.obj is src


# ---------------------------------------------------------------------------
# DTXFragmenter.fragment() - multi-fragment path
# ---------------------------------------------------------------------------


class TestFragmentMultiFragment:
    async def test_two_fragments_exact_split(self):
        """Payload is exactly 2 * MAX_FRAGMENT_SIZE."""
        total = MAX_FRAGMENT_SIZE * 2
        src = bytearray(list(range(256)) * (total // 256))
        frags = await _collect(DTXFragmenter.fragment(memoryview(src)))

        # header-only + 2 body fragments
        assert len(frags) == 3
        assert frags[0].index == 0
        assert frags[0].count == 3
        assert frags[0].data_size == total
        assert len(frags[0].payload) == 0

        assert frags[1].index == 1
        assert frags[1].data_size == MAX_FRAGMENT_SIZE
        assert frags[2].index == 2
        assert frags[2].data_size == MAX_FRAGMENT_SIZE

        # Reassemble and verify
        assembled = bytes(frags[1].payload) + bytes(frags[2].payload)
        assert assembled == bytes(src)

    async def test_three_fragments_partial_last(self):
        """Payload is 2.5 * MAX_FRAGMENT_SIZE → 3 body fragments, last partial."""
        total = int(MAX_FRAGMENT_SIZE * 2.5)
        src = bytearray(b"\xab" * total)
        frags = await _collect(DTXFragmenter.fragment(memoryview(src)))

        # header-only + 3 body fragments
        assert len(frags) == 4
        body_frags = frags[1:]
        assembled = b"".join(bytes(f.payload) for f in body_frags)
        assert assembled == bytes(src)

    async def test_large_payload_roundtrip(self):
        """Fragment then reassemble a large payload through DTXFragmenter."""
        total = MAX_FRAGMENT_SIZE * 3 + 12345
        src = bytearray(i % 251 for i in range(total))
        frags = await _collect(DTXFragmenter.fragment(memoryview(src)))

        body_frags = frags[1:]
        first = frags[0]

        assembler = DTXFragmenter(
            first,
            current_buffered=0,
            max_buffered_size=MAX_MESSAGE_SIZE,
        )
        for frag in body_frags:
            done = assembler.add(frag)

        assert done is True
        result, _ = assembler.assemble()
        assert bytes(result) == bytes(src)

    async def test_zero_copy_body_fragments_single_source(self):
        """Each body fragment from a single large source chunk must be a sub-view."""
        total = MAX_FRAGMENT_SIZE * 2 + 1000
        src = bytearray(b"\xcc" * total)
        frags = await _collect(DTXFragmenter.fragment(memoryview(src)))

        body_frags = frags[1:]
        for frag in body_frags:
            # Every body fragment must share the same backing bytearray as src.
            assert frag.payload.obj is src, (
                f"Fragment {frag.index} payload not zero-copy: payload.obj={id(frag.payload.obj):#x}, src={id(src):#x}"
            )

    async def test_boundary_fragment_allocates_new_buffer(self):
        """When a fragment straddles a source-chunk boundary, a new bytearray is allocated."""
        # Split the source into two half-sized chunks so the first body
        # fragment boundary falls exactly at the chunk boundary — BUT then
        # use unequal splits to force a boundary-crossing fragment.
        half = MAX_FRAGMENT_SIZE // 2
        chunk_a = bytearray(b"\xaa" * (half + 1))  # slightly over half
        chunk_b = bytearray(b"\xbb" * (MAX_FRAGMENT_SIZE - 1))  # fills the rest of the second body frag

        frags = await _collect(DTXFragmenter.fragment(memoryview(chunk_a), memoryview(chunk_b)))
        assert len(frags) == 3  # header + 2 body

        body_frags = frags[1:]
        assembled = b"".join(bytes(f.payload) for f in body_frags)
        assert assembled == bytes(chunk_a) + bytes(chunk_b)

        # At least one body fragment must NOT share obj with either source chunk
        # (the one that crossed the chunk boundary gets its own allocation).
        boundary_frag = body_frags[0]  # the first body fragment straddles the split
        assert boundary_frag.payload.obj is not chunk_a
        assert boundary_frag.payload.obj is not chunk_b

    async def test_exceeds_max_message_size_raises(self):
        oversized = bytearray(MAX_MESSAGE_SIZE + 1)
        with pytest.raises(DTXProtocolError, match="MAX_MESSAGE_SIZE"):
            await _collect(DTXFragmenter.fragment(memoryview(oversized)))

    async def test_roundtrip_out_of_order_assembly(self):
        """Fragment a payload, then feed body fragments to assembler in reverse order."""
        total = MAX_FRAGMENT_SIZE * 2 + 7777
        src = bytearray(i % 199 for i in range(total))
        frags = await _collect(DTXFragmenter.fragment(memoryview(src)))

        first_frag = frags[0]
        body_frags = frags[1:]  # original order: index 1, 2, 3...

        assembler = DTXFragmenter(
            first_frag,
            current_buffered=0,
            max_buffered_size=MAX_MESSAGE_SIZE,
        )

        # Feed in reverse order.
        for frag in reversed(body_frags):
            done = assembler.add(frag)

        assert done is True
        result, _ = assembler.assemble()
        assert bytes(result) == bytes(src)


# ---------------------------------------------------------------------------
# DTXFragmenter.fragment() - multiple source chunks, multi-fragment
# ---------------------------------------------------------------------------


class TestFragmentMultipleSourceChunks:
    async def test_many_small_chunks(self):
        """Many small chunks (each 1 byte) spanning multiple fragments."""
        chunk_count = MAX_FRAGMENT_SIZE * 2 + 500
        chunks = [memoryview(bytearray([i % 256])) for i in range(chunk_count)]
        frags = await _collect(DTXFragmenter.fragment(*chunks))

        body_frags = frags[1:]
        assembled = b"".join(bytes(f.payload) for f in body_frags)
        assert len(assembled) == chunk_count
        assert assembled == bytes(i % 256 for i in range(chunk_count))

    async def test_two_chunks_same_size_as_one_fragment(self):
        """Two chunks each of half MAX_FRAGMENT_SIZE → exactly one body fragment."""
        half = MAX_FRAGMENT_SIZE // 2
        a = bytearray(b"\x01" * half)
        b_ = bytearray(b"\x02" * half)
        frags = await _collect(DTXFragmenter.fragment(memoryview(a), memoryview(b_)))
        # total == MAX_FRAGMENT_SIZE → single fragment (no multi-fragment path)
        assert len(frags) == 1
        assert bytes(frags[0].payload) == bytes(a) + bytes(b_)

    async def test_index_sequence_is_contiguous(self):
        """Fragment indices must be 0, 1, 2, … for any payload size."""
        total = MAX_FRAGMENT_SIZE * 4 + 3
        src = bytearray(total)
        frags = await _collect(DTXFragmenter.fragment(memoryview(src)))
        for expected_idx, frag in enumerate(frags):
            assert frag.index == expected_idx

    async def test_count_is_consistent(self):
        """All fragments in a batch must carry the same count value."""
        total = MAX_FRAGMENT_SIZE * 3 + 1
        src = bytearray(total)
        frags = await _collect(DTXFragmenter.fragment(memoryview(src)))
        counts = {f.count for f in frags}
        assert len(counts) == 1
        assert frags[0].count == len(frags)

    async def test_total_body_size_matches_declared(self):
        """Sum of body fragment data_sizes must equal first fragment's data_size."""
        total = MAX_FRAGMENT_SIZE * 2 + 987
        src = bytearray(total)
        frags = await _collect(DTXFragmenter.fragment(memoryview(src)))
        declared = frags[0].data_size
        body_total = sum(f.data_size for f in frags[1:])
        assert body_total == declared
        assert declared == total



import json
from io import StringIO

import pytest
import pytest_asyncio

from pymobiledevice3.cli.developer.dvt.sysmon.process import (
    ProcessSelectionMode,
    _process_sort_key,
    _select_process_from_snapshot,
    iter_processes,
    sysmon_process_monitor_threshold_task,
    sysmon_process_single_task,
)
from pymobiledevice3.services.dvt.instruments.sysmontap import Sysmontap


@pytest_asyncio.fixture
async def process_snapshot(dvt) -> list[dict]:
    async with await Sysmontap.create(dvt) as sysmon:
        async for process_snapshot in iter_processes(sysmon):
            return process_snapshot

    pytest.fail("failed to collect an initialized process snapshot")


@pytest.mark.asyncio
async def test_iter_processes_yields_process_dicts(process_snapshot) -> None:
    assert len(process_snapshot) > 0
    assert isinstance(process_snapshot[0], dict)
    assert process_snapshot[0].get("pid")


@pytest.mark.asyncio
async def test_select_process_from_snapshot_filters_by_existing_pid(process_snapshot) -> None:
    process = process_snapshot[0]

    selected_process = _select_process_from_snapshot(
        process_snapshot, {"pid": [str(process["pid"])]}, ProcessSelectionMode.FIRST
    )

    assert selected_process.get("pid") == process.get("pid")


@pytest.mark.asyncio
async def test_select_process_from_snapshot_first_and_last_follow_sorted_order(
    process_snapshot,
) -> None:
    if len(process_snapshot) < 2:
        pytest.skip("requires at least two processes in the current snapshot")

    sorted_processes = sorted(process_snapshot, key=_process_sort_key)

    assert _select_process_from_snapshot(process_snapshot, {}, ProcessSelectionMode.FIRST) == sorted_processes[0]
    assert _select_process_from_snapshot(process_snapshot, {}, ProcessSelectionMode.LAST) == sorted_processes[-1]


@pytest.mark.asyncio
async def test_sysmon_process_single_task_writes_json_to_buffer(service_provider) -> None:
    out = StringIO()

    await sysmon_process_single_task(service_provider, keys=["pid", "name"], out=out)

    result = json.loads(out.getvalue())
    assert len(result) > 0
    assert result[0].get("pid")
    assert result[0].get("name")


@pytest.mark.asyncio
async def test_sysmon_process_monitor_threshold_task_writes_jsonl_to_buffer(service_provider) -> None:
    out = StringIO()

    await sysmon_process_monitor_threshold_task(service_provider, threshold=0.0, duration=500, keys=["pid"], out=out)

    lines = [line for line in out.getvalue().splitlines() if line.strip()]
    assert len(lines) > 0
    first_record = json.loads(lines[0])
    assert first_record.get("pid")



#!/usr/bin/env python3
import dataclasses
import plistlib
import struct
import tempfile
import typing
from datetime import datetime
from enum import IntEnum
from pathlib import Path
from tarfile import TarFile

from pymobiledevice3.exceptions import ConnectionTerminatedError, PyMobileDevice3Exception
from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.lockdown_service_provider import LockdownServiceProvider
from pymobiledevice3.services.lockdown_service import LockdownService
from pymobiledevice3.utils import try_decode

CHUNK_SIZE = 4096
TIME_FORMAT = "%H:%M:%S"
SYSLOG_LINE_SPLITTER = "\n\x00"


class SyslogLogLevel(IntEnum):
    NOTICE = 0x00
    INFO = 0x01
    DEBUG = 0x02

    # deducted from console-app
    USER_ACTION = 0x03
    ERROR = 0x10
    FAULT = 0x11


@dataclasses.dataclass
class SyslogLabel:
    category: str
    subsystem: str


@dataclasses.dataclass
class SyslogEntry:
    pid: int
    timestamp: datetime
    level: SyslogLogLevel
    image_name: str
    image_offset: int
    filename: str
    message: str
    label: typing.Optional[SyslogLabel] = None


def parse_syslog_entry(data: bytes) -> SyslogEntry:
    """
    Parse a syslog entry from binary data.

    :param data: Raw binary data
    :return: SyslogEntry
    """
    offset = 0

    # Skip first 9 bytes
    offset += 9

    # Parse pid (4 bytes, little-endian unsigned int)
    pid = struct.unpack("<I", data[offset : offset + 4])[0]
    offset += 4

    # Skip 42 bytes
    offset += 42

    # Parse timestamp
    seconds = struct.unpack("<I", data[offset : offset + 4])[0]
    offset += 4
    offset += 4  # Skip 4 bytes
    microseconds = struct.unpack("<I", data[offset : offset + 4])[0]
    offset += 4
    timestamp = datetime.fromtimestamp(seconds + (microseconds / 1000000))

    # Skip 1 byte
    offset += 1

    # Parse level (1 byte)
    level = data[offset]
    offset += 1

    # Skip 38 bytes
    offset += 38

    # Parse image_name_size (2 bytes, little-endian unsigned short)
    image_name_size = struct.unpack("<H", data[offset : offset + 2])[0]
    offset += 2

    # Parse message_size (2 bytes, little-endian unsigned short)
    message_size = struct.unpack("<H", data[offset : offset + 2])[0]
    offset += 2

    # Skip 2 bytes
    offset += 2

    # Parse sender_image_offset (4 bytes, little-endian unsigned int)
    sender_image_offset = struct.unpack("<I", data[offset : offset + 4])[0]
    offset += 4

    # Parse subsystem_size (4 bytes, little-endian unsigned int)
    subsystem_size = struct.unpack("<I", data[offset : offset + 4])[0]
    offset += 4

    # Parse category_size (4 bytes, little-endian unsigned int)
    category_size = struct.unpack("<I", data[offset : offset + 4])[0]
    offset += 4

    # Skip 4 bytes
    offset += 4

    # Parse filename (null-terminated)
    filename_end = data.find(b"\x00", offset)
    filename = try_decode(data[offset:filename_end])
    offset = filename_end + 1

    # Parse image_name
    image_name = try_decode(data[offset : offset + image_name_size - 1])
    offset += image_name_size

    # Parse message
    message = try_decode(data[offset : offset + message_size - 1])
    offset += message_size

    # Parse label (optional)
    label = None
    if subsystem_size > 0 and category_size > 0:
        subsystem = try_decode(data[offset : offset + subsystem_size - 1])
        offset += subsystem_size
        category = try_decode(data[offset : offset + category_size - 1])
        offset += category_size
        label = SyslogLabel(subsystem=subsystem, category=category)

    return SyslogEntry(
        pid=pid,
        timestamp=timestamp,
        level=SyslogLogLevel(level),
        image_name=image_name,
        image_offset=sender_image_offset,
        filename=filename,
        message=message,
        label=label,
    )


class OsTraceService(LockdownService):
    """
    Provides API for the following operations:
    * Show process list (process name and pid)
    * Stream syslog lines in binary form with optional filtering by pid.
    * Get old stored syslog archive in PAX format (can be extracted using `pax -r < filename`).
        * Archive contain the contents are the `/var/db/diagnostics` directory
    """

    SERVICE_NAME = "com.apple.os_trace_relay"
    RSD_SERVICE_NAME = "com.apple.os_trace_relay.shim.remote"

    def __init__(self, lockdown: LockdownServiceProvider):
        if isinstance(lockdown, LockdownClient):
            super().__init__(lockdown, self.SERVICE_NAME)
        else:
            super().__init__(lockdown, self.RSD_SERVICE_NAME)

    async def get_pid_list(self):
        await self.connect()
        assert self.service is not None
        await self.service.send_plist({"Request": "PidList"})

        # ignore first received unknown byte
        await self.service.recvall(1)

        response = await self.service.recv_prefixed()
        return plistlib.loads(response)

    async def create_archive(
        self,
        out: typing.IO,
        size_limit: typing.Optional[int] = None,
        age_limit: typing.Optional[int] = None,
        start_time: typing.Optional[int] = None,
    ):
        request = {"Request": "CreateArchive"}

        if size_limit is not None:
            request.update({"SizeLimit": size_limit})

        if age_limit is not None:
            request.update({"AgeLimit": age_limit})

        if start_time is not None:
            request.update({"StartTime": start_time})

        await self.connect()
        assert self.service is not None
        await self.service.send_plist(request)

        assert (await self.service.recvall(1))[0] == 1

        assert plistlib.loads(await self.service.recv_prefixed()).get("Status") == "RequestSuccessful", "Invalid status"

        while True:
            try:
                assert (await self.service.recvall(1))[0] == 3, "invalid magic"
            except ConnectionTerminatedError:
                break
            out.write(await self.service.recv_prefixed(endianity="<"))

    async def collect(
        self,
        out: str,
        size_limit: typing.Optional[int] = None,
        age_limit: typing.Optional[int] = None,
        start_time: typing.Optional[int] = None,
    ) -> None:
        """
        Collect the system logs into a .logarchive that can be viewed later with tools such as log or Console.

        :param out: output file name
        :param size_limit: maximum size in bytes of logarchive
        :param age_limit: maximum age in days
        :param start_time: start time of logarchive in unix timestamp
        """
        with tempfile.TemporaryDirectory() as temp_dir:
            file = Path(temp_dir) / "foo.tar"
            with open(file, "wb") as f:
                await self.create_archive(f, size_limit=size_limit, age_limit=age_limit, start_time=start_time)
            TarFile(file).extractall(out)

    async def syslog(self, pid=-1) -> typing.AsyncGenerator[SyslogEntry, None]:
        await self.connect()
        assert self.service is not None
        await self.service.send_plist({
            "Request": "StartActivity",
            "MessageFilter": 65535,
            "Pid": pid,
            "StreamFlags": 60,
        })

        (length_length,) = struct.unpack("<I", await self.service.recvall(4))
        length = int((await self.service.recvall(length_length))[::-1].hex(), 16)
        response = plistlib.loads(await self.service.recvall(length))

        if response.get("Status") != "RequestSuccessful":
            raise PyMobileDevice3Exception(f"got invalid response: {response}")

        while True:
            assert await self.service.recvall(1) == b"\x02"
            (length,) = struct.unpack("<I", await self.service.recvall(4))
            line = await self.service.recvall(length)
            yield parse_syslog_entry(line)
